# encoding: utf-8

import re
from itertools import izip

from crf.data.dataset import DataSet, divide_data, divide_in_two

CASES = ["nom", "gen", "dat", "acc", "inst", "loc", "voc"]
CASE_REGEX = re.compile("(^|:)(%s)($|:)" % '|'.join(CASES))

class Segment(object):
    
    def __init__(self, id, orth="?", base="?", pos="?", morph="?", nps=False):
        self._id = id
        self._nps = nps
        self._orth = orth
        self._base = base
        self._pos = pos
        self._morph = morph
        self._case = self._get_case(morph)
    
    @property
    def id(self):
        return self._id
    
    @property
    def nps(self):
        return self._nps
    
    @property
    def orth(self):
        return self._orth
    
    @property
    def base(self):
        return self._base
    
    @property
    def pos(self):
        return self._pos

    def _get_case(self, morph):
        match = CASE_REGEX.search(morph)
        assert match != None if self._pos in ["subst", "depr", "ger",
                "ppron12", "ppron3", "num", "numcol", "adj", "pact",
                "ppas", "prep", "siebie"] else match == None
        if match != None:
            return match.group(2)
        return None

    @property
    def case(self):
        return self._case
    
    @property
    def morph(self):
        return self._morph
    
    def __eq__(self, other):
        return (self.id, self.orth, self.base, self.pos, self.morph) \
            == (other.id, other.orth, other.base, other.pos, other.morph)
    
    def __hash__(self):
        return hash(self.id)
    
    def __repr__(self):
        return ('Segment(id=%s, orth=%s, interp=%s)' \
                % (self.id,
                   self.orth, 
                   ':'.join([self.base, self.pos, self.morph]))) \
                .encode('utf-8', 'ignore')

class DummySegment(Segment):

    def __init__(self, id):
        Segment.__init__(self, id, "?", "?", "?", "?")

    def __repr__(self):
        return "DummySegment(id=%s)" % self.id

class NamedEntity(object):
    
    def __init__(self, 
                 id, 
                 type, 
                 subtype, 
                 ptrs, 
                 base=None,
                 derivType=None, 
                 derivedFrom=None):
        self._id = id
        self._type = type
        self._subtype = subtype
        self._ptrs = ptrs
        self._base = base
        self._derivType = derivType
        self._derivedFrom = derivedFrom
    
    @property
    def id(self):
        return self._id
    
    @id.setter
    def id(self, id):
        self._id = id
    
    @property
    def type(self):
        return self._type
    
    @property
    def subtype(self):
        return self._subtype
    
    @property
    def ptrs(self):
        return self._ptrs
    
    @property
    def derivType(self):
        return self._derivType
    
    @property
    def derivedFrom(self):
        return self._derivedFrom
    
    @property
    def base(self):
        return self._base
    
    def get_segs(self):
        res = []
        for ptr in self.ptrs:
            if isinstance(ptr, Segment):
                res.append(ptr)
            else: 
                res.extend(ptr.get_segs())
        return res

    def get_child_segs(self):
        return [ptr for ptr in self.ptrs if isinstance(ptr, Segment)]
    
    def get_child_names(self):
        return [ptr for ptr in self.ptrs if isinstance(ptr, NamedEntity)]
    
    def get_descendant_names_and_self(self):
        return [self] + self.get_descendant_names()
    
    def get_descendant_names(self):
        res = []
        for name in self.get_child_names():
            res.extend(name.get_descendant_names_and_self())
        return res
    
    def get_label(self):
        if self.subtype != None:
            out = self.type + "." + self.subtype
        else:
            out = self.type
        if self.derivType != None:
            out = out + "@" + self.derivType
        return out

    def _get_orth(self):
        out = []
        for ptr in self.ptrs:
            if isinstance(ptr, NamedEntity):
                out.append(ptr._get_orth())
            else:
                if not ptr.nps:
                    out.append(' ')
                out.append(ptr.orth)
        return ''.join(out)
    
    @property
    def orth(self):
        return self._get_orth().strip()
    
    def __eq__(self, other):
        return self is other
    
    def __hash__(self):
        return hash(self.id)
    
    def _ptrs_str(self):
        res = ['[']
        for ptr in self.ptrs:
            res.append(ptr.id)
            if len(self.ptrs) > 1:
                res.append(', ')
        res.append(']')
        return ''.join(res)
    
    def __repr__(self):
        out = ['NamedEntity(']
        out.append('id=%s, type=%s, ' % (self.id, self.type))
        if self.subtype is not None:
            out.append('subtype=%s, ' % self.subtype)
        if self.base is not None:
            out.append('base=%s, ' % self.base)
        if self.derivType is not None:
            out.append('derivType=%s, ' % self.derivType)
        if self.derivedFrom is not None:
            out.append('derivedFrom=%s, ' % self.derivedFrom)
        out.append('ptrs=%s)' % self._ptrs_str())
        return ''.join(out)

    def __str__(self):
        val = ""
        for seg in sorted(self.get_segs(), key=lambda seg: seg.id):
            val += str(seg) + "\n"
        return val.strip()

def flatten(l):
    result = []
    for subl in l:
        result.extend(subl)
    return result

def join_names(names1, names2, discard_nested=True):
    """Try to put NEs from names2 list to names1 list; return joined lists."""
    if discard_nested != True:
        raise NotImplementedError
    coverage = set(flatten(ne.get_segs() for ne in names1))
    result = names1
    for ne in names2:
        if set(ne.get_segs()) & coverage == set():
            result.append(ne)
    return result

class NerDataSet(DataSet):

    """CRF DataSet extended with NEs for each sentence."""

    def __init__(self, rdata, *args, **kwargs):
        self.raw_data = [raw_sent for raw_sent, _ in rdata]
        DataSet.__init__(self, [sent for _, sent in rdata], *args, **kwargs)

    def iterator(self):
        return ((raw_sent, sent)
                for raw_sent, sent
                in izip(self.raw_data, self.data))

    def divide(self, n, stat=(lambda x: 1)):
        """Divide dataset data to n parts."""
        return divide_data(self.iterator(), self.dirs, n, stat=stat)

    def divide_in_two(self, r):
        return divide_in_two(self.iterator(), self.dirs, r)

def filter_raw(data):
    return [sent for (raw_sent, sent) in data]



"""
Module provides methods for counting statistics both on the
level of labels and on the level on NEs.

Methods work for translated CRF data.
"""

from crf.data import translate as translate
import crf.stats

from ner.errors import errors
from ner.agreement import Agreement
from ner.data import DummySegment
from ner.labels import label2tuple, decode

def _retranslate_sent(sent, rev_alph):
    sent, labels = sent
    return (translate.sentence(sent, rev_alph),
            translate.labels(labels, rev_alph))

def label_stats(data, rev_alph, ner):
    data = (_retranslate_sent(sent, rev_alph) for sent in data)
    return crf.stats.stats(data, ner, ner.alphabet.void_label)

#def name_stats(data, ner, atlas=None, predicate=None, **kwds):
#    if atlas != None:
#        raise NotImplementedError
#    data = (_retranslate_sent(sent, ner.rev_alphabet) for sent in data)
#    agr = Agreement(**kwds)
#    tagged, good, agrees = 0, 0, 0
#    for (sent, labels) in data:
#        segs = [DummySegment(str(i)) for i in range(len(sent))]
#        labels = map(label2tuple, labels)
#        names_tagged = ner.recognize_named_entities(segs, features=sent)
#        names_good = decode(labels, segs)
#        (_tagged, _good, _ok) = agr.get_agreement(names_tagged, names_good,
#                predicate=predicate)
#        tagged += _tagged
#        good += _good
#        agrees += _ok
#
#    prec = rec = 0.0
#    if tagged > 0:
#        prec = float(agrees) / tagged
#    if good > 0:
#        rec = float(agrees) / good
#    if prec == 0.0 and rec == 0.0:
#        F = 0.0
#    else:
#        F = (2*prec*rec) / (prec + rec)
#    # return (tagged, good, agrees, prec, rec, F)
#    return (("Tagged", tagged),
#            ("Good", good),
#            ("Agrees", agrees),
#            ("Precision", prec),
#            ("Recall", rec),
#            ("F-measure", F))

def precision(tagged, good, agrees):
    if tagged > 0:
        return float(agrees) / tagged
    return 0.0

def recall(tagged, good, agrees):
    if good > 0:
        return float(agrees) / good
    return 0.0

def F(tagged, good, agrees):
    prec = precision(tagged, good, agrees)
    rec = recall(tagged, good, agrees)
    if prec == 0.0 or rec == 0.0:
        return 0.0
    else:
        return (2*prec*rec) / (prec + rec)

#def name_stats(data, ner, predicates=[None], **kwds):
#    data = (_retranslate_sent(sent, ner.rev_alphabet) for sent in data)
#    agr = Agreement(**kwds)
#
#    stats = []
#    for _ in predicates:
#        stats.append((0, 0, 0))    # tagged, good, agrees
#
#    for (sent, labels) in data:
#        segs = [DummySegment(str(i)) for i in range(len(sent))]
#        labels = map(label2tuple, labels)
#        names_tagged = ner.recognize_named_entities(segs, features=sent)
#        names_good = decode(labels, segs)
#
#        for i, pred in enumerate(predicates):
#            (tagged, good, ok) = stats[i]
#            (_tagged, _good, _ok) = agr.get_agreement(names_tagged,
#                    names_good, predicate=pred)
#            stats[i] = (tagged + _tagged, good + _good, ok + _ok)
#
#    return stats

def name_stats(data, ner, predicates=[None], **kwds):
    # data = (_retranslate_sent(sent, ner.rev_alphabet) for sent in data)
    agr = Agreement(**kwds)

    stats = []
    for _ in predicates:
        stats.append((0, 0, 0))    # tagged, good, agrees

    for (segs, names_good), _ in data:
        # segs = [DummySegment(str(i)) for i in range(len(sent))]
        # labels = map(label2tuple, labels)
        # names_tagged = ner.recognize_named_entities(segs, features=sent)
        names_tagged = ner.recognize_named_entities(segs)
        # names_good = decode(labels, segs)

        for i, pred in enumerate(predicates):
            (tagged, good, ok) = stats[i]
            (_tagged, _good, _ok) = agr.get_agreement(names_tagged,
                    names_good, predicate=pred)
            stats[i] = (tagged + _tagged, good + _good, ok + _ok)

    return stats

def _old_count_names(name):
    return 1 + sum(_old_count_names(child) for child in name.get_child_names())

def old_count_names(data, rev_alphabet=None):
    """Number of names in data."""
    value = 0
    obset = set()
    for _, (sent, labels) in data:
        segs = [DummySegment(str(i)) for i in range(len(sent))]
        if rev_alphabet is not None:
            labels = translate.labels(labels, rev_alphabet)
        labels = map(label2tuple, labels)
        names = decode(labels, segs)
        for name in names:
            value += _old_count_names(name)
    return value

#def _orth(seginfo):
#    (singles, _) = seginfo
#    for pref in ["4_", "2_"]:
#        for ob in singles:
#            if ob.startswith(pref):
#                return ob.replace(pref, "")
#    return "X"

#def _tagged_data(data, ner):
#    for (sent, labels) in data:
#        segs = [DummySegment(str(i)) for i in range(len(sent))]
#        labels = map(label2tuple, labels)
#        names_tagged = ner.recognize_named_entities(segs, features=sent)
#        names_good = decode(labels, segs)
#        yield map(_orth, sent), names_tagged, names_good

def _count_names(name):
    return len(name.get_descendant_names_and_self())

def count_names(data):
    return sum(_count_names(name)
            for ((_, names), _) in data
            for name in names)

def _tagged_data(data, ner):
    # for (sent, labels) in data:
    for (segs, names_good), _ in data:
        # segs = [DummySegment(str(i)) for i in range(len(sent))]
        # labels = map(label2tuple, labels)
        # names_tagged = ner.recognize_named_entities(segs, features=sent)
        names_tagged = ner.recognize_named_entities(segs)
        # names_good = decode(labels, segs)
        yield segs, names_tagged, names_good
        # yield map(_orth, sent), names_tagged, names_good

def count_errors(data, ner):
    # data = (_retranslate_sent(sent, ner.rev_alphabet) for sent in data)
    return errors(_tagged_data(data, ner))



#-*- coding: utf-8 -*-

import sys
import os
import re
import string

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from ner.nerecognizer import NERecognizer
from ner.data import Segment
from ner.features import sent2features

# Word or punctuation character
word_regex = re.compile('([^%s]+|[%s])' %
        ( re.escape(string.whitespace + string.punctuation)
        , re.escape(string.punctuation)))
def words(line):
    return [(m.group(), m.span()) for m in word_regex.finditer(line)]

class SegmentStruct(Segment):

    def __init__(self, span, *args, **kwargs):
        self.span = span
        Segment.__init__(self, *args, **kwargs)

def line2sent(line):
    return [ SegmentStruct(span, id=str(i), orth=orth)
             for i, (orth, span) in enumerate(words(line)) ]

def name2str(name):
    segs = sorted(name.get_segs(), key=lambda seg: seg.span)
    npss = [True] + [prev.span[1] == curr.span[0]
                     for (prev, curr) in zip(segs, segs[1:])]
    segstr = "".join(
            ("" if nps else " ") + seg.orth
            for (seg, nps) in zip(segs, npss))
    p = min(seg.span[0] for seg in segs)
    q = max(seg.span[1] for seg in segs)
    span_str = "(" + str(p) + ", " + str(q) + ")"
    return (".".join(filter(lambda x: x is not None,
        [name.type, name.subtype, name.derivType]))
        + " : " + segstr + " " + span_str)

def print_names(names):
    for root_name in names:
        for name in root_name.get_descendant_names_and_self():
            print name2str(name).encode('utf-8')

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog MODEL
        
        NER program, which reads text from stdin
        (each sentence in separate line) and
        writes results to stdout.""")

    (options, args) = optparser.parse_args()
    if len(args) != 1:
        optparser.print_help()
        sys.exit(0)

    model_name = args[0]
    # print "Loading model..."
    ner = NERecognizer.load(model_name)
    # print "Done\n"

    # for line in sys.stdin:
    while True:
        line = sys.stdin.readline()
        if line == "":
            break
        sent = line2sent(line.decode('utf-8').strip())
        names = ner.recognize_named_entities(sent)
        print_names(names)
        print



# encoding: utf-8

# cdef extern from "libcrf/connsbuf.h":
cdef extern from "connsbuf.h":
    ctypedef struct C_Conns:
        int** conns
        int* csizes
        int xn

cdef class Conns:
    cdef C_Conns conns

# cdef extern from "libcrf/model.h":
cdef extern from "model.h":
    ctypedef struct C_Model:
        int pn
        int* features
        double* values
        int xn
        int yn
        int dummy_id

cdef class Model:
    cdef C_Model model
    cdef object fmap
    cdef Conns conns



#!/usr/bin/env python

import sys
import os

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from crf.data import dataset
from crf.data import rawdata
from crf.stats import stats
from crf.train.sgd import sgd
from crf.tager import Tager
from crf.api import CRF
from crf.model import Model

def file_lines(file_name):
    return map(lambda x: x.strip(), open(file_name).readlines())

class CallBack:

    def __init__(self, train, evl, tager, void_id, every=10):
        self.train = train
        self.evl = evl
        self.tager = tager
        self.void_id = void_id
        self.every = every
        self.prev_iter = 0.0

    def __call__(self, iter):
#        if i % 50 == 0:
#            print "=== Train part ==="
#            prec, rec, F = stats(self.train, self.tager, self.void_id)
#            print "Precision:", prec
#            print "Recall:", rec
#            print "F-measure:", F
#
        if int(iter / self.every) > int(self.prev_iter / self.every):
            print "=== Iteration: %s ===" % round(iter, 1)
            prec, rec, F = stats(self.evl, self.tager, self.void_id)
            print "Precision:", prec
            print "Recall:", rec
            print "F-measure:", F
        self.prev_iter = iter

if __name__ == '__main__':
    from optparse import OptionParser
    optparser = OptionParser(
        usage="""%prog TRAIN_FILE MODEL_OUTPUT_FILE,
        
        * Read data from TRAIN_FILE,
        * Divide data into two parts: train and eval,
        * Train CRF model on the train part,
        * Compute presision, recall and F-measure on both parts 
          with respect to the CRF model,
        * Save CRF model in the MODEL_OUTPUT_FILE.""")
	
    optparser.add_option("-r", "--relative_eval_size", type="float",
            default=0.1, dest="eval_size",
            help="relative (to the data size) size of the eval part")
#    optparser.add_option("-i", "--init-model", dest="init_model_dir",
#            metavar="DIR", help="directory with initial model")
    optparser.add_option("-d", "--dirs_file", dest="dirs_file",
            metavar="FILE", help="file with names of directories corresponding"
            " to sentences in the data; when supplied, data will be divided"
            " with respect to directories")
#     optparser.add_option("-s", "--save-freq", type="int",
#             dest="save_freq", help="Saving model frequency.")
#     optparser.add_option("-m", "--max-train-f", type="float", default=100.0,
#             dest="max_train_f", help="Stop when F-measure"
#                 " on train file is greater then max-train-f.")
    optparser.add_option("-b", "--batch_size", type="int",
            default=30, dest="batch_size",
            help="batch size for the SGD method")
    optparser.add_option("-n", "--max_iter_num", type="int", default=100,
            dest="max_iter_num", help="number of SGD method iterations")
    optparser.add_option("-t", "--threads", type="int", default=1,
            dest="threads", help="number of threads in parallel computation")
    optparser.add_option("--regvar", type="float", dest="regvar",
            help="regularization variance parameter, see SGD specification")
    optparser.add_option("--scale0", type="float", default=0.01,
            dest="scale0", help="scale0 parameter, see SGD specification")
    optparser.add_option("--tau", type="float", default=10.0,
            dest="tau", help="tau parameter, see SGD specification")
    optparser.add_option("--void_label", type="str", dest="void_label",
            default="O", help="label representing void information")
    optparser.add_option("--dummy_label", type="str", dest="dummy_label",
            default="DUMMY", help="label corresponding to the posision in"
            " a sentence before the first word")
    optparser.add_option("-v", "--verbose", default=False,
            action="store_true", dest="verbose")
    optparser.add_option("--pre-train", default=False,
            action="store_true", dest="pre_train",
            help="pre-train model with smaller scale0.")

    (options, args) = optparser.parse_args()
    if len(args) != 2:
        optparser.print_help()
        sys.exit()
    file_name, model_output = args

    rdata = rawdata.from_file(file_name)
    if options.dirs_file != None:
        dirs = file_lines(options.dirs_file)
    else:
        dirs = None
    ds = dataset.DataSet(rdata,
            void_label=unicode(options.void_label),
            dummy_label=unicode(options.dummy_label),
            dirs=dirs)
    train, evl = ds.divide_in_two(options.eval_size)

    model = Model(dataset.features_in_data(train))
    tager = Tager(model, ds.obtypes())
    void_id = ds.alphabet.label(unicode(options.void_label))

    callback = CallBack(train, evl, tager, void_id, every=1.0)
    if options.pre_train:
        sgd(model, train,
                iter_num=options.max_iter_num / 10.0,
                batch_size=options.batch_size,
                scale0=options.scale0 / 10.0,
                tau=options.tau, workers=options.threads,
                callback=callback, verbose=False)
    sgd(model, train, callback=callback,
            iter_num=options.max_iter_num,
            batch_size=options.batch_size,
            scale0=options.scale0,
            tau=options.tau,
            workers=options.threads,
            verbose=options.verbose)
#    sgd(model, train, max_iter_num=options.max_iter_num,
#            batch_size=options.batch_size, scale0=options.scale0,
#            tau=options.tau, workers=options.threads,
#            callback=callback, verbose=options.verbose)

    print "Saving model in %s..." % model_output
    crf = CRF(model, ds.alphabet, ds.obtypes())
    crf.save(model_output)
    print "Done"
    print ""

    prec, rec, F = stats(train, tager, void_id)
    print " === Train part ==="
    print " Precision:", prec
    print " Recall:", rec
    print " F-measure:", F

    prec, rec, F = stats(evl, tager, void_id)
    print " === Eval part ==="
    print " Precision:", prec
    print " Recall:", rec
    print " F-measure:", F



# encoding: utf-8

from libc.stdlib cimport malloc, free

from crf.model cimport Model, Conns
from crf.data.connsbuf cimport ConnsBuf
from crf.data.auxdata cimport AuxData
from crf.data.sentence cimport Sentence

cdef extern from "sentence.h":
    ctypedef struct C_Sentence:
        pass

cdef extern from "conns.h":
    ctypedef struct C_Model:
        pass

cdef extern from "conns.h":
    ctypedef struct C_Conns:
        pass

cdef extern from "connsbuf.h":
    ctypedef struct C_ConnsBuf:
        pass

cdef extern from "utils.h":
    void tag_sent(C_Sentence* sent, C_Model* model, C_Conns* conns,
            C_ConnsBuf* connsBuf, int* labels) nogil

def tag(Sentence sent, Model model, ConnsBuf connsbuf):
    cdef int* labels = <int*>malloc(len(sent) * sizeof(int))
    if labels is NULL:
        raise MemoryError
    with nogil:
        tag_sent(&sent.sent, &model.model, &model.conns.conns,
                connsbuf.cbuf, labels)
    value = [labels[i] for i in range(len(sent))]
    free(labels)
    return value

class Tager:

    def __init__(self, model, obtypes):
        self.model = model
        self.connsbuf = ConnsBuf(model.yn, obtypes)

    def tag(self, Sentence sent):
        return tag(sent, self.model, self.connsbuf)



import gzip
import os

from xml.sax.saxutils import escape

__all__ = ["TEIWriter", "write"]
    
def _writeProlog(out):
    out.write("""<?xml version="1.0" encoding="UTF-8"?>
<teiCorpus xmlns:xi="http://www.w3.org/2001/XInclude" xmlns="http://www.tei-c.org/ns/1.0">
 <xi:include href="NKJP_1M_header.xml"/>
 <TEI>
  <xi:include href="header.xml"/>
  <text xml:lang="pl">
   <body>""")
 
def _writeEpilogue(out):
    out.write("""</body>
  </text>
 </TEI>
</teiCorpus>
""")

def _ptr2target(id):
    if id.startswith('morph'):
        return 'ann_morphosyntax.xml#%s' % id
    else:
        return id

def _writeNE(out, ne):
#    print "NE:", ne.id, ne.type, ne.subtype,\
#            ne.base, ne.derivType, ne.derivedFrom
    out.write("""
     <seg xml:id="%s">
      <fs type="named">
       <f name="type">
        <symbol value="%s"/>
       </f>""" % (ne.id, ne.type))
    if ne.subtype is not None:
        out.write("""
       <f name="subtype">
        <symbol value="%s"/>
       </f>""" % ne.subtype)
    if ne.derivType is not None:
        out.write("""
       <f name="derived">
        <fs type="derivation">
         <f name="derivType">
          <symbol value="%s"/>
         </f>""" % ne.derivType)
        if ne.derivedFrom is not None:
            out.write("""
         <f name="derivedFrom">
          <string>%s</string>
         </f>""" % escape(derivedFrom).encode('utf-8'))
        out.write("""
        </fs>
       </f>""")

#        derivedFrom = "" if ne.derivedFrom is None else ne.derivedFrom 
#        out.write("""
#       <f name="derived">
#        <fs type="derivation">
#         <f name="derivType">
#          <symbol value="%s"/>
#         </f>
#         <f name="derivedFrom">
#          <string>%s</string>
#         </f>
#        </fs>
#       </f>""" % (ne.derivType, escape(derivedFrom).encode('utf-8')))
    out.write("""
       <f name="orth">
        <string>%s</string>
       </f>""" % escape(ne.orth).encode('utf-8'))
    out.write("""
      </fs>""")
    for ptr in ne.ptrs:
        out.write("""
      <ptr target="%s"/>""" % _ptr2target(ptr.id))
    out.write("""
     </seg>""")

def _correct_ne_ids(named_sent_id, all_names):
    for i, ne in enumerate(all_names, start=1):
        ne.id = '%s_n%d' % (named_sent_id, i)
        # print ne.id

def _writeSent(out, named_sent_id, sent_id, names_list):
    all_names = reduce(lambda names1, names2: names1 + names2,
                       map(lambda name: name.get_descendant_names_and_self(), 
                           names_list))
    out.write("""
    <s xml:id="%s" corresp="ann_morphosyntax.xml#%s">""" 
                % (named_sent_id, sent_id))
    _correct_ne_ids(named_sent_id, all_names)
    for ne in all_names:
        _writeNE(out, ne)
    out.write("""
    </s>""")

def _writeEmptySent(out, named_sent_id, sent_id):
    out.write("""
    <s xml:id="%s" corresp="ann_morphosyntax.xml#%s"/>""" 
                % (named_sent_id, sent_id))

def _writeParagraph(out, par_id, contents):
    out.write("""
   <p xml:id="%s" corresp="ann_morphosyntax.xml#%s">"""
             % (par_id.replace('morph', 'named'), par_id))
    for sent_id, sent_names in contents:
        named_sent_id = sent_id.replace('morph', 'named')
        if sent_names != []:
            _writeSent(out, named_sent_id, sent_id, sent_names)
        else:
            _writeEmptySent(out, named_sent_id, sent_id)
    #print 'written:', par_id.replace('morph', 'named')
    out.write("""
   </p>""")
    
def write(out, paragraphs):
    _writeProlog(out)
    for par_id, contents in paragraphs:
        _writeParagraph(out, par_id, contents)
    _writeEpilogue(out)



# encoding: utf-8

from collections import defaultdict

"""
Module provides methods for error analisys.
"""

__all__ = ["errors"]

#FP = "false positive"
#FN = "false negative"
#OK = "ok"

def netype(ne):
    info = [ne.type, ne.subtype, ne.derivType]
    return ":".join(filter(None, info))

def map_ne(root):
    """Map recursively each ne to a pair (type, segments) and
    return created list."""
    all_nes = []
    root_segs = set() 
    for child in root.get_child_names():
        nes, segs = map_ne(child)
        all_nes.extend(nes)
        root_segs.update(segs)
    # root_segs.update(int(seg.id) for seg in root.get_child_segs())
    root_segs.update(root.get_child_segs())
    all_nes.append((netype(root), root_segs))
    return all_nes, root_segs

def map_nes(nes):
    """Map each ne in the nes structure to a pair (type, segments)."""
    value = []
    for ne in nes:
        value.extend(map_ne(ne)[0])
    return value

def dist(ne1, ne2):
    f1, s1 = ne1
    f2, s2 = ne2
    if len(s1 & s2) > 0:
        return len(s1.symmetric_difference(s2)) + int(f1 != f2)
    else:
        return float('inf')

def minarg(f, args):
    result = None
    value = float('inf')
    for arg in args:
        fval = f(arg)
        if fval < value:
            value = fval
            result = arg
    return result

def _tostr(segs):
    return "".join(seg.orth if seg.nps else " " + seg.orth
            for seg in segs).strip()

def stats(l1, l2, ok, no_corresp, pairs):
    fn_errors, subst_errors = [], []
    for ne1 in l1:
        (f1, s1) = ne1
        ne2 = minarg(lambda ne2: dist(ne1, ne2), l2)
        if ne2 is None:
            no_corresp[f1] += 1
            fn_errors.append((f1, s1))
            continue
        (f2, s2) = ne2
        if dist(ne1, ne2) == 0:
            ok[f1] += 1
        elif s1 & s2 != set():
            pairs[(f1, f2)] += 1
            subst_errors.append(((f1, s1), (f2, s2)))
        else:
            no_corresp[f1] += 1
            fn_errors.append((f1, s1))
    return fn_errors, subst_errors

def segs2str(segs):
    return "".join(("" if seg.nps else " ") + seg.orth
            for seg in segs).strip()

def sent_errors(segs, recognized, expected, tp, fp, fn, substs, pairs):
    """
    Collect and return list of errors in one sentence.
    
    :params:
    --------
    tp :
        True positives.
    fp :
        False positives.
    fn :
        False negatives.
    substs :
        Substitutions; dictionary entry for (T1, T2) key means the number
        of occurences, that instead expected T1 named entity, T2 has been
        recognized.
    pairs :
        Pairs, meaning inverse to this of substs.
    """
    # sent = " ".join(orths).encode('utf-8')
    sent = segs2str(segs).encode('utf-8')
    _recognized = map_nes(recognized)
    _expected = map_nes(expected)
    (fp_errors, pair_errors) = stats(_recognized, _expected, tp, fp, pairs)
    (fn_errors, subst_errors) = stats(_expected, _recognized,
            defaultdict(int), fn, substs)
    # fn_errors = [((f, " ".join(orths[i] for i in sorted(s))), sent)
    fn_errors = [((f, segs2str(sorted(s))), sent)
            for (f, s) in fn_errors]
    # fp_errors = [((f, " ".join(orths[i] for i in sorted(s))), sent)
    fp_errors = [((f, segs2str(sorted(s))), sent)
            for (f, s) in fp_errors]
    subst_errors = [(
        (f1, segs2str(sorted(s1))),
        (f2, segs2str(sorted(s2))),
        sent) for (f1, s1), (f2, s2) in subst_errors]
    pair_errors = [(
        (f1, segs2str(sorted(s1))),
        (f2, segs2str(sorted(s2))),
        sent) for (f1, s1), (f2, s2) in pair_errors]
    return (fn_errors, fp_errors, subst_errors, pair_errors)

def errors(data):
    """Perform error analisys given data.
    
    :params:
    --------
    data : iterable
        List of pairs (R, E), where R is a list of recognized NEs, while
        E is a list of NEs present in sentences within the gold corpus.
    """
    tp = defaultdict(int)
    fp = defaultdict(int)
    fn = defaultdict(int)
    substs = defaultdict(int)
    pairs = defaultdict(int)
    fn_errors, fp_errors, subst_errors, pair_errors = [], [], [], []
    for segs, recognized, expected in data:
        sent_fns, sent_fps, sent_substs, sent_pairs = sent_errors(segs,
                recognized, expected, tp, fp, fn, substs, pairs)
        fn_errors += sent_fns
        fp_errors += sent_fps
        subst_errors += sent_substs
        pair_errors += sent_pairs
    return tp, fp, fn, substs, pairs, fn_errors, fp_errors,\
            subst_errors, pair_errors



# encoding: utf-8

from libc.stdlib cimport malloc, free

from crf.array cimport DArray
from crf.model cimport Model
from crf.data.connsbuf cimport ConnsBuf
from crf.data.auxdata cimport AuxData
from crf.data.sentence cimport Sentence

cdef extern from "sentence.h":
    ctypedef struct C_Sentence:
        pass

cdef extern from "model.h":
    ctypedef struct C_Model:
        int pn
        int* features
        double* values
        int xn
        int yn
        int dummy_id

cdef extern from "conns.h":
    ctypedef struct C_Conns:
        pass

cdef extern from "connsbuf.h":
    ctypedef struct C_ConnsBuf:
        pass

cdef extern from "utils.h":
    double logZ(C_Sentence* sent, C_Model* model, C_Conns* conns,
            C_ConnsBuf* connsbuf, double* B) nogil

def count_z_stats(batch, Model model, ConnsBuf connsbuf):
    cdef double logz
    cdef Sentence sent
    z_stats = []
    for sent, _ in batch:
        with nogil:
            logz = logZ(&sent.sent, &model.model, &model.conns.conns,
                    connsbuf.cbuf, NULL)
        z_stats.append(logz)
    return z_stats

def add_expected_numbers(Sentence sent, DArray numbers_wrapper, Model model,
        ConnsBuf connsbuf):
    """Add expected numbers of features in the sentence."""
    cdef int k
    cdef double* numbers = numbers_wrapper.array
    cdef C_Model* _model = &model.model
    cdef double logz
    with nogil:
        logz = logZ(&sent.sent, _model, &model.conns.conns,
                connsbuf.cbuf, numbers)
    return logz

def expected_numbers(batch, DArray numbers_wrapper, Model model,
        ConnsBuf connsbuf):
    """Count expected numbers of features in the batch of sentences."""
    cdef int k
    cdef double* numbers = numbers_wrapper.array
    cdef C_Model* _model = &model.model
    with nogil:
        for k from 0 <= k < _model.pn:
            numbers[k] = 0.0
    cdef Sentence sent
    cdef double logz
    z_stats = []
    for sent, _ in batch:
        with nogil:
            logz = logZ(&sent.sent, _model, &model.conns.conns,
                    connsbuf.cbuf, numbers)
        z_stats.append(logz)
    return z_stats

def count_gradient(batch, DArray gradient_wrapper, Model model, 
        AuxData auxdata, ConnsBuf connsbuf, double i_var2, double coef):
    """
    :param coef:
        Shoudl be equall to |batch size| / |data size|.
    """
    cdef int fn, i
    cdef double val
    expected_numbers(batch, gradient_wrapper, model, auxdata, connsbuf)
    cdef double* gradient = gradient_wrapper.array
    cdef C_Model* _model = &model.model
    # ANOTHER VERSION: auxdata = AuxData(batch, model)
    with nogil:
        for i from 0 <= i < _model.pn:
            val = _model.values[i]
            fn = auxdata.fnum[i]
            # gradient[i] = gradient[i] - fn + val*i_var2*coef
            gradient[i] = gradient[i] - (fn + val*i_var2)*coef

#def count_gradient(batch, GradientWrapper gradient_wrapper, Model model,
#        AuxData auxdata, ConnsBuf connsbuf, double i_var2,
#        double coef, count_cll=True, verbose=False):
#    """
#    :param coef:
#        Shoudl be equall to |batch size| / |data size|.
#    """
#    cdef int k
#    cdef double* gradient = gradient_wrapper.gradient
#    cdef C_Model* _model = &model.model
#
#    with nogil:
#        for k from 0 <= k < _model.pn:
#            gradient[k] = 0.0
#        
#    # OTHER VERSION:
#    # auxdata = AuxData(batch, model)
#
#    cdef Sentence sent
#    cdef double logz
#    z_stats = []
#    for sent, _ in batch:
#        with nogil:
#            logz = logZ(&sent.sent, _model, &model.conns.conns,
#                    connsbuf.cbuf, gradient)
#        z_stats.append(logz)
#
#    cdef int fn
#    cdef double val
#    with nogil:
#        for i from 0 <= i < _model.pn:
#            val = _model.values[i]
#            fn = auxdata.fnum[i]
#            # gradient[i] = gradient[i] - fn + val*i_var2*coef
#            gradient[i] = gradient[i] - (fn + val*i_var2)*coef
#
#    cllval = 0.0
#    if count_cll:
#        cllval = model.cll(batch, z_stats, coef, i_var2)
#        if verbose:
#            print "Conditional Log Likelihood:", cllval
#    return cllval



# encoding: utf-8

cdef class ConnsBuf:

    def __cinit__(self):
        self.cbuf = NULL

    def __init__(self, yn, obstype_num):
        self.cbuf = newConnsBuf(yn, obstype_num)
        if self.cbuf is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self.cbuf is not NULL:
            freeConnsBuf(self.cbuf)



# encoding: utf-8

from libc.stdlib cimport malloc, free

cdef class Sentence:

    def __cinit__(self):
        self.sent.length = 0
        self.sent.singles = NULL
        self.sent.snum = NULL 
        self.sent.pairs = NULL
        self.sent.pnum = NULL 
        self.sent.snum = NULL 

    def _alloc(self, size, snum, pnum):
        self.sent.singles = <int**>malloc(size * sizeof(int*))
        if self.sent.singles is NULL:
            raise MemoryError
        self.sent.pairs = <int**>malloc(size * sizeof(int*))
        if self.sent.pairs is NULL:
            raise MemoryError
        self.sent.snum = <int*>malloc(size * sizeof(int))
        if self.sent.snum is NULL:
            raise MemoryError
        self.sent.pnum = <int*>malloc(size * sizeof(int))
        if self.sent.pnum is NULL:
            raise MemoryError
        for i in range(size):
            self.sent.singles[i] = NULL
            self.sent.pairs[i] = NULL
        for i in range(size):
            self.sent.singles[i] = <int*>malloc(snum[i] * sizeof(int))
            if self.sent.singles[i] is NULL:
                raise MemoryError
            self.sent.pairs[i] = <int*>malloc(pnum[i] * sizeof(int))
            if self.sent.pairs[i] is NULL:
                raise MemoryError

    def _dealloc(self):
        if self.sent.snum is not NULL:
            free(self.sent.snum)
        if self.sent.pnum is not NULL:
            free(self.sent.pnum)
        if self.sent.singles is not NULL:
            for i in range(self.sent.length):
                if self.sent.singles[i] is not NULL:
                    free(self.sent.singles[i])
            free(self.sent.singles)
        if self.sent.pairs is not NULL:
            for i in range(self.sent.length):
                if self.sent.pairs[i] is not NULL:
                    free(self.sent.pairs[i])
            free(self.sent.pairs)

    def __dealloc__(self):
        self._dealloc()

    def __init__(self, sent):
        self._dealloc()
        self.sent.length = len(sent)
        # self._alloc(len(sent), map(len, sent))
        self._alloc(len(sent),
                [len(ss) for (ss, _) in sent],
                [len(ps) for (_, ps) in sent])
        for i, (singles, pairs) in enumerate(sent):
            self.sent.snum[i] = len(singles)
            self.sent.pnum[i] = len(pairs)
            for j, x in enumerate(singles):
                self.sent.singles[i][j] = x
            for j, x in enumerate(pairs):
                self.sent.pairs[i][j] = x

    def __getitem__(self, i):
        if i < 0 or i >= self.sent.length:
            raise IndexError
        return ([self.sent.singles[i][j] for j in range(self.sent.snum[i])],
                [self.sent.pairs[i][j] for j in range(self.sent.pnum[i])])

    def __len__(self):
        return self.sent.length



#-*- coding: utf-8 -*-

import sys
import os
import time
import re
from threading import Thread
from Queue import Queue
import string

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from ner.nerecognizer import NERecognizer
from ner.data import Segment
from ner.features import sent2features

# Word or punctuation character
word_regex = re.compile('([^%s]+|[%s])' %
        ( re.escape(string.whitespace + string.punctuation)
        , re.escape(string.punctuation)))
def words(line):
    return [(m.group(), m.span()) for m in word_regex.finditer(line)]

class SegmentStruct(Segment):

    def __init__(self, span, *args, **kwargs):
        self.span = span
        Segment.__init__(self, *args, **kwargs)

class NERService(object):

    def __init__(self, ner, input, output):
        self.ner = ner
        self.input = input
        self.output = output
    
    def group_input(self):
        result = []
        while True:
            line = self.input.get()
            if line is None:
                break
            yield [SegmentStruct(span, id=str(i), orth=orth)
                for i, (orth, span) in enumerate(words(line))]

    def name2str(self, name):
        segs = sorted(name.get_segs(), key=lambda seg: seg.span)
        npss = [True] + [prev.span[1] == curr.span[0]
                         for (prev, curr) in zip(segs, segs[1:])]
        segstr = "".join(
                ("" if nps else " ") + seg.orth
                for (seg, nps) in zip(segs, npss))
        p = min(seg.span[0] for seg in segs)
        q = max(seg.span[1] for seg in segs)
        span_str = "(" + str(p) + ", " + str(q) + ")"
        return (".".join(filter(lambda x: x is not None,
            [name.type, name.subtype, name.derivType]))
            + " : " + segstr + " " + span_str)
    
    def output_names(self, root_names):
        names = [ name
                  for root_name in root_names
                  for name in root_name.get_descendant_names_and_self() ]
        self.output.put(len(names))
        for name in names:
            namestr = self.name2str(name)
            self.output.put(namestr)
    
    def run(self):
        for sent in self.group_input():
#            features = sent2features(sent, self.ner.features)
#            print >> self.output, "FEATURES:"
#            for elem in features:
#                print >> self.output, " ".join(elem).encode('utf-8')
            names = self.ner.recognize_named_entities(sent)
            self.output_names(names)

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog MODEL
        
        Run NER service using given model.""")

    (options, args) = optparser.parse_args()
    if len(args) != 1:
        optparser.print_help()
        sys.exit(0)

    model_name = args[0]
    print "Loading model..."
    ner = NERecognizer.load(model_name)
    print "Done\n"

    input = Queue()
    output = Queue()
    service = NERService(ner, input=input, output=output)
    thread = Thread(target=service.run, args=())
    thread.start()

    try:
        while True:
            line = raw_input("> ")
            input.put(line.decode('utf-8'))
            k = output.get()
            for _ in range(k):
                print output.get().encode('utf-8')
    except KeyboardInterrupt:
        input.put(None)

    thread.join()



# encoding: utf-8

cdef extern from "../libcrf/sentence.h":
    ctypedef struct C_Sentence:
        int length
        int** singles   # single observations
        int* snum       # number of singles obs.
        int** pairs     # pair observations
        int* pnum       # number of pair obs.

cdef class Sentence:
    cdef C_Sentence sent



#-*- coding: utf-8 -*-

class Agreement:
    def __init__(self, 
                 ignore_base=True, 
                 ignore_derivType=True,
                 ignore_derivedFrom=True):
        self.ignore_base = ignore_base
        self.ignore_derivType = ignore_derivType
        self.ignore_derivedFrom = ignore_derivedFrom

    def _get_names_num(self, names, predicate):
        res = 0
        for name in names:
            res += self._get_names_num(name.get_child_names(), predicate)
            # if categories is None or name.type in categories:
            if predicate is None or predicate(name):
                res += 1
        return res
    
    def _name2tuple(self, name):
        res = []
        res.append(frozenset([seg.id for seg in name.get_segs()]))
        res.append(name.type)
        res.append(name.subtype)
        if not self.ignore_base:
            res.append(name.base)
        if not self.ignore_derivType:
            res.append(name.derivType)
        if not self.ignore_derivedFrom:
            res.append(name.derivedFrom)
        
        return tuple(res)
    
    def _get_all_names(self, names, predicate):
        res = []
        for name in names:
            # if categories is None or name.type in categories:
            if predicate is None or predicate(name):
                res.append(name)
            res.extend(self._get_all_names(name.get_child_names(), predicate))
        return res
    
    def _names2tuples_multiset(self, names, predicate):
        res = {}
        for name in self._get_all_names(names, predicate):
            name_tuple = self._name2tuple(name)
            res.setdefault(name_tuple, 0)
            res[name_tuple] += 1
        return res

    def _get_names_ok(self, names_test, names_good, predicate):
        res = 0
        ne_test_tuples = self._names2tuples_multiset(names_test, 
                                                     predicate)
        ne_good_tuples = self._names2tuples_multiset(names_good, 
                                                     predicate)
        for test_tuple, _ in ne_test_tuples.iteritems():
            good_count = ne_good_tuples.get(test_tuple, 0)
            res += good_count
        return res
    
    def get_agreement(self, names_test, names_good, predicate=None):
        count_test = self._get_names_num(names_test, predicate)
        count_good = self._get_names_num(names_good, predicate)
        count_ok = self._get_names_ok(names_test, 
                                      names_good, 
                                      predicate)
        return (count_test, count_good, count_ok)
    



# encoding: utf-8

def filternone(l):
    return filter(lambda x: x != None, l)

def sentence(sent, alph):
    return [(filternone([alph.observation(x) for x in singles]),
        filternone([alph.observation(x) for x in pairs]))
        for singles, pairs in sent]

def labels(labels, alph):
    return [alph.label(x) for x in labels]



# encoding: utf-8

def from_file(file_name):
    """Create raw data from file."""
    # WARNING: IT WILL NOT WORK PROPERLY, IF THERE
    # ARE WHITESPACE CHARACTERS IN OBSERVARTIONS
    # OR LABELS !
    data = []
    words = []
    labels = []

    file = open(file_name)
    for line in file:
        # line = line.strip()
        if len(line.strip()) == 0:
            data.append((words, labels))
            words = []
            labels = []
        else:
            singles, pairs, label = (x.strip() for x in line.split(" | "))
            # line = line.split()
            word = ([x.decode('utf-8') for x in singles.split()],
                    [x.decode('utf-8') for x in pairs.split()])
            # obs_list = [obs.decode('utf-8') for obs in line[:-1]]
            # label = line[-1].decode('utf-8')
            label = label.decode('utf-8')
            words.append(word)
            labels.append(label)

    file.close()
    if len(words) != 0:
        raise Exception("no empty line on the end of the %s file." % file_name)
    return data



# encoding: utf-8

"""
Representation of CRF model.
"""

import os
import gzip
import math
import const
from collections import defaultdict

from libc.stdlib cimport malloc, free

include "conns.pxi"

class ParamsIter:
    def __init__(self, model):
        self.model = model
        self.idx = -1
    def next(self):
        self.idx += 1
        if self.idx >= self.model.pn:
            raise StopIteration
        return (self.model.feature(self.idx), self.model.value(self.idx))

def load(filename=None, fileobj=None):
    """Load CRF model saved using custom format."""
    if not fileobj:
        inp = gzip.open(filename, 'rb')
    else:
        inp = fileobj
    features = []
    values = []
    for line in inp:
        f1, f2, f3, val = line.strip().split()
        features.append(map(int, (f1, f2, f3)))
        values.append(float(val))
    cdef Model model = Model(features)
    for i, v in zip(range(model.pn), values):
        model.model.values[i] = v
    return model

cdef class Model:

    """
    Class representing CRF model parameters.

    Representation of CRF model consists of two elements:
    * Alphabet, or map, from strings to observation/label indices.
    * Model parameters, that is list of (feature, value) pairs.
    The Model class is responsible for the representation of the
    second element from the list above.

    Let X be a set of observation indices and Y a set of label indices.
    The following assumptions has to be satisfied:
    * X = {0, 1, ..., |X| - 1}
    * Y = {0, 1, ..., |Y| - 1}

    Model features are of one of the forms:
    * Simple featuer -- (-1, y, x).
    * Compound feature -- (py, y, x),
    """

    def __cinit__(self):
        self.model.pn = 0
        self.model.xn = 0
        self.model.yn = 0
        self.model.dummy_id = 0
        self.model.features = NULL
        self.model.values = NULL
        self.conns = None
        self.fmap = {}

    def _dealloc(self):
        self.fmap.clear()
        if self.model.features is not NULL:
            free(self.model.features)
        if self.model.values is not NULL:
            free(self.model.values)

    def __dealloc__(self):
        self._dealloc()

    def _create_map(self):
        self.fmap.clear()
        for i in range(self.pn):
            feature = self.feature(i)
            self.fmap[feature] = i

    def _observations(self, features):
        value = set()
        for (py, y, x) in features:
            value.add(x)
        return value

    def _labels(self, features):
        value = set()
        for (py, y, x) in features:
            if py != -1:
                value.add(py)
            value.add(y)
        return value

    def __init__(self, features, values=None):
        if values is None:
            values = [0.0] * len(features)
        assert len(features) == len(values)

        self._dealloc()

        # Hard-codding dummy label id. Note, that this has to be
        # consistent with alphabet ! (see alphabet.py)
        self.model.dummy_id = const.dummy_id
        self.model.pn = len(features)

        oset = self._observations(features)
        self.model.xn = max(oset) + 1
        assert oset <= set(range(self.model.xn))
        lset = self._labels(features)
        self.model.yn = max(lset) + 1
        assert lset <= set(range(self.model.yn))

        self.model.features = <int*>malloc(3 * self.pn * sizeof(int))
        if self.model.features is NULL:
            raise MemoryError
        self.model.values = <double*>malloc(self.pn * sizeof(double))
        if self.model.values is NULL:
            raise MemoryError

        for i, (py, y, x) in enumerate(features):
            k = i * 3
            self.model.features[k] = py
            self.model.features[k + 1] = y
            self.model.features[k + 2] = x
        for i, v in enumerate(values):
            self.model.values[i] = v

        self._create_map()
        self.conns = Conns(self)

    property pn:
        def __get__(self):
            return self.model.pn

    property xn:
        def __get__(self):
            return self.model.xn

    property yn:
        def __get__(self):
            return self.model.yn

    property dummy_id:
        def __get__(self):
            return self.model.dummy_id

    def feature(self, i):
        """i-th feature"""
        if i < 0 or i >= self.pn:
            raise ValueError
        value = (self.model.features[i*3],
                self.model.features[i*3 + 1],
                self.model.features[i*3 + 2])
        return value

    def value(self, i):
        """value for the i-th feature"""
        if i < 0 or i >= self.pn:
            raise ValueError
        return self.model.values[i]

    def feature_id(self, feature):
        return self.fmap[feature]

    def __getitem__(self, feature):
        i = self.fmap[feature]
        return self.model.values[i]

    def get(self, feature, default=None):
        try:
            return self[feature]
        except:
            return default

    def __iter__(self):
        return ParamsIter(self)

    def save(self, filename=None, fileobj=None):
        """Save CRF model using custom format."""
        if not fileobj:
            out = gzip.open(filename, 'wb')
        else:
            out = fileobj

        for feature, value in self:
            print >> out, ' '.join(map(str, list(feature) + [value]))

        if not fileobj:
            out.close()

    def log_phi(self, py, y, word):
        value = 0.0
        singles, pairs = word
        for x in singles:
            feature = (-1, y, x)
            value += self.get(feature, 0.0)
        for x in pairs:
            feature = (py, y, x)
            value += self.get(feature, 0.0)
        return value

    # TODO: Count Z constant within the method.
    def prob(self, sentence, labels, log_z, log=True):
        """Probability of the sentence and labels with respect to the model.
        
        :param log_z:
            Logarithm of a Z contstant related to the sentence.
            See CRF specification for more details.
        """
        value = 0.0

        # prev_y = self.alphabet.label(self.alphabet.dummyLabel)
        py = self.dummy_id
        for word, y in zip(sentence, labels):
            value += self.log_phi(py, y, word);
            py = y
            
        value -= log_z
        #if value > 0.:
        #	print 'value > 0 !'
        #	print 'value: ' + str(value)
        #	print 'logZx: ' + str(logZx)
        if not log:
            return math.exp(value)
        return value

    def cll(self, data, z_stats, double i_var2, double coef=1.0):
        """
        :param coef:
            Relative (to the size of entire training set) size of the data.
        """
        cdef double value = 0.0
        cdef int i
        cdef double v

        fl = False
        for (i, (sent, labels)) in enumerate(data):
            logp = self.prob(sent, labels, z_stats[i])
            if logp > 0.0:
                #logp = 0.
                fl = True
            value += logp
            
        coef = coef * 0.5 * i_var2
        for i from 0 <= i < self.pn:
            v = self.model.values[i]
            value -= (v*v) * coef

        if fl:
            print 'problems with numbers precision'
        return value

    def norm_2(self):
        sum2 = 0.0
        for i in range(self.pn):
            v = self.model.values[i]
            sum2 += v*v
        return math.sqrt(sum2)

    def translated(self, from_alph, to_alph):
        """
        Translate model to a new alphabet.
        
        Parameters, which cannot be translated to the new alphabet,
        will be discarded.
        """
        rev_from = from_alph.reversed()
        features, values = [], []
        for feature, value in self:
            feature = rev_from.translate_feature(feature)
            feature = to_alph.translate_feature(feature)
            if feature != None:
                features.append(feature)
                values.append(value)
        return Model(features, values=values)



#!/usr/bin/env python

import sys
import os
import copy

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from crf.model import Model
from crf.data.dataset import divide_in_two, labels_in_data, features_in_data
from crf.train.sgd import sgd

from ner.nerecognizer import NERecognizer
from ner.data import NerDataSet, filter_raw
from ner.labels import decode, label2tuple

from tei.corpus import read_tei_corpus
from tei.stats import name_stats, count_names
import tei.stats as stats_mod

def file_lines(filename):
    return map(lambda x: x.strip(), open(filename).readlines())

def _adorn_stats(stats):
    tagged, good, agrees = stats
    prec = stats_mod.precision(*stats)
    rec = stats_mod.recall(*stats)
    F = stats_mod.F(*stats)
    return (("Tagged", tagged),
            ("Good", good),
            ("Agrees", agrees),
            ("Precision", prec),
            ("Recall", rec),
            ("F-measure", F))

def get_overall(data, ner):
    stats = name_stats(data, ner, ignore_derivType=False)[0]
    adorned = _adorn_stats(stats)
    return {"(1) overall": adorned}

def type_eq(netype):
    return lambda ne: ne.type == netype

def nkjp_summary_groups():
    groups = [["persName"], ["geogName", "placeName"], ["orgName"],
            ["date", "time"], ["relAdj", "persDeriv"]] 
    result = []
    result.append((["persName"],
        lambda name: name.type in ["persName"]))
    result.append((["geogName", "placeName"],
        lambda name: name.type in ["geogName", "placeName"]))
    result.append((["orgName"],
        lambda name: name.type in ["orgName"]))
    result.append((["date", "time"],
        lambda name: name.type in ["date", "time"]))
    result.append((["relAdj", "persDeriv"],
        lambda name: name.derivType in ["relAdj", "persDeriv"]))
    result.append((["overall"], None))
    return result

def netypes_in_dataset(dataset):
    netypes = set()
    for (_, names), _ in dataset:
        for root in names:
            for ne in root.get_descendant_names_and_self():
                netypes.add(ne.type)
    return sorted(netypes)

def get_summary(data, ner, groups):
#    preds = [lambda name: name.type in copy.deepcopy(group)
#            for group in groups] + [None]
    preds = [pred for _, pred in groups]
    groups = [group for group, _ in groups]
    stat_list = name_stats(data, ner,
            predicates=preds, ignore_derivType=False)
    summary = {}
    for i, (stats, group) in enumerate(zip(stat_list, groups)):
        # groups + [["overall"]])):
        adorned_stats = _adorn_stats(stats)
        summary[("(%s) " % (i + 1)) + " and ".join(group)] = adorned_stats 
    return summary

def print_summary(summary):
    for category, adorned_stats in sorted(summary.iteritems()):
        print category + ":"
        for statname, val in adorned_stats:
            print statname + ":", val
        print ""

class CallBack:

    """Callback for SGD method."""

    def __init__(self, train, evl, ner, every=1):
        """
        :params:
        --------
        every : int
            Compute stats on eval part each $every iteration. 
        """
        self.train = train
        self.evl = evl
        self.ner = ner
        self.every = every
        self.prev_iter = 0.0

    def __call__(self, iter):
        if int(iter / self.every) > int(self.prev_iter / self.every):
            print "[%s]" % round(iter, 1),
            print "F(eval) =",
            stats = name_stats(self.evl, self.ner,
                    ignore_derivType=False)[0]
            adorned = _adorn_stats(stats)
            print dict(adorned)["F-measure"]
        self.prev_iter = iter

def train(data, evl, dataset, schema, ner=None, sgd_kwargs={},
        eval_every=10.0, generate_absent=False): 
    """
    Train NER tool with respect to the given data.

    :params:
    --------
    data :
        List of (sentence, labels) pairs.
    dataset : NerDataSet
        Source of both data and evl parts.
    ner : NERecognizer
        If not None, ner will be used as an initial NER model.
    """
    if ner is not None:
        model = ner.model.translated(
                from_alph=ner.alphabet,
                to_alph=dataset.alphabet)
    else:
        model = Model(features_in_data(filter_raw(data),
            generate_absent=generate_absent))

    print "Number of model parameters:", model.pn

    ner = NERecognizer(model, dataset.alphabet, schema)
    callback = CallBack(data, evl, ner, every=eval_every)
    sgd(model, filter_raw(data), callback=callback, **sgd_kwargs)
    return ner

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog,
        
        Extract data from TEI corpus, divide it to training part and
        evaluation part and perform CRF model training process.""")
	
    optparser.add_option("--tei-corpus", metavar="DIR", dest="tei_corpus",
            help="Use TEI corpus as data source.")
    optparser.add_option("--schema", metavar="FILE",
            dest="schema", help="File with feature schema.")
    optparser.add_option("--generate-absent-features", default=False,
            action="store_true", dest="generate_absent",
            help="Generate features absent in data.")
    optparser.add_option("--model-out", metavar="FILE",
            dest="model_out", help="Resultant model output file.")
    optparser.add_option("--model-in", metavar="FILE", dest="model_in",
            help="Start training from initial model.")
    optparser.add_option("--ann-named", type="string",
            dest="ann_named", default="ann_named.xml",
            help="Names of input ann_named files.")
    optparser.add_option("--nkjp-summary", default=False,
            action="store_true", dest="nkjp_summary",
            help="Print summary adapted to NKJP.")
    optparser.add_option("--data-out", metavar="FILE",
            dest="data_out", help="Data output file (for debugging purpose).")
    optparser.add_option("--select", type="float", default=1.0,
            dest="select", help="Select subcorpus of given relative size.")
    optparser.add_option("--black-listed", metavar="FILE",
            dest="black_listed", help="Black listed corpus directories.")
    optparser.add_option("--relative-eval-size", type="float",
            default=0.1, dest="eval_size",
            help="Relative (to the data size) size of the eval part.")
    optparser.add_option("--evaluate-every", type="int",
            default=10, dest="eval_every",
            help="Compute stats on evaluation part every * interations.")
    optparser.add_option("--batch-size", type="int",
            default=30, dest="batch_size",
            help="Batch size in SGD method.")
    optparser.add_option("--iter-num", type="int", default=100,
            dest="iter_num", help="Number of SGD method iterations.")
    optparser.add_option("--threads", type="int", default=1, dest="threads",
            help="Number of threads in parallel computation.")
    optparser.add_option("--regvar", type="float", dest="regvar",
            help="Regularization variance parameter, see SGD specification.")
    optparser.add_option("--scale0", type="float", default=0.01,
            dest="scale0", help="Scale0 parameter, see SGD specification.")
    optparser.add_option("--tau", type="float", default=10.0,
            dest="tau", help="Tau parameter, see SGD specification.")
    optparser.add_option("--pre-train", default=False,
            action="store_true", dest="pre_train",
            help="Pre-train model on input data with predefined," +
                    " small scale0 value.")
    optparser.add_option("--void-label", type="str", dest="void_label",
            default="O", help="Label representing void information.")
    optparser.add_option("--dummy-label", type="str", dest="dummy_label",
            default="DUMMY", help="Label corresponding to the posision in"
            " a sentence before the first word.")
#    optparser.add_option("--verbose_level", type="int",
#            default=1, dest="verbose_level",
#            help="0: no output, 1: ...")

    (options, args) = optparser.parse_args()
    if options.tei_corpus == None:
        optparser.print_help()
        print "\n--tei-corpus option is mandatory"
        sys.exit(0)
    if options.schema == None and options.model_in == None:
        optparser.print_help()
        print "\n--schema or --model-in option should be supplied"
        sys.exit(0)
    if options.schema != None and options.model_in != None:
        print ("WARNING: schema will be ignored, using model-in" +
                " schema instead")

    ner = None
    schema = None
    if options.model_in:
        print "Loading model from %s..." % options.model_in
        ner = NERecognizer.load(options.model_in)
        schema = ner.schema
    else:
        schema = file_lines(options.schema)

    if options.black_listed != None:
        black_listed = file_lines(options.black_listed)
    else:
        black_listed = []

    print "Reading data..."
    raw_data, dirs = read_tei_corpus(options.tei_corpus, schema,
            black_listed, options.select, ann_named=options.ann_named)
    dataset = NerDataSet(raw_data, void_label=unicode(options.void_label),
            dummy_label=unicode(options.dummy_label), dirs=dirs)
    print "Reading done\n"

#    print "Data statistics:"
#    print "Number of sentences:", len(dataset)
#    print "Number of segments:", sum(len(sent) for sent, labels in dataset)
#    print "Number of NEs:", count_names(dataset)
#    print ""
    if options.data_out:
        dataset.print_data(options.data_out)

    train_part, eval_part = dataset.divide_in_two(options.eval_size)
    print "Train part statistics:"
    print "Number of sentences:", len(train_part)
    print "Number of segments:", sum(len(segs) for (segs, _), _ in train_part)
    print "Number of labels:", len(labels_in_data(filter_raw(train_part)))
    print "Number of NEs:", count_names(train_part)
    print ""
    if len(eval_part) > 0:
        print "Evaluation part statistics:"
        print "Number of sentences:", len(eval_part)
        print "Number of segments:", sum(len(segs)
                for (segs, _), _ in eval_part)
        print "Number of NEs:", count_names(eval_part)
        print ""

    if options.pre_train:
        print "Pre-training model..."
        ner = train(train_part, eval_part, dataset, schema, ner=ner, 
                eval_every=options.eval_every,
                generate_absent=options.generate_absent,
                sgd_kwargs={
                    "iter_num": 10,
                    "batch_size": options.batch_size,
                    "tau": options.tau,
                    "scale0": 0.01,
                    "workers": options.threads,
                    "verbose": False
                    })
        print "Pre-training done\n"

    print "Training model..."
    # ner = train(train_part, eval_part, options.iter_num, dataset, schema,
    #         ner=ner, scale0=options.scale0, workers=options.threads)
    ner = train(train_part, eval_part, dataset, schema, ner=ner, 
            eval_every=options.eval_every,
            generate_absent=options.generate_absent,
            sgd_kwargs={
                "iter_num": options.iter_num,
                "batch_size": options.batch_size,
                "tau": options.tau,
                "scale0": options.scale0,
                "workers": options.threads,
                "verbose": False
                })

    if options.model_out:
        print ""
        print "Saving model in %s..." % options.model_out
        ner.save(options.model_out)
        print "Done"

    if options.nkjp_summary:
        summary_groups = nkjp_summary_groups()
    else:
        summary_groups = [([netype], type_eq(netype))
                for netype in netypes_in_dataset(dataset.iterator())]
        summary_groups += [(["overall"], None)]

    print ""
    print "====== Summary ======"
    print ""
    print "[Train part]"
    print ""
    # train_summary = get_overall(train_part, ner)
    train_summary = get_summary(train_part, ner, summary_groups)
    print_summary(train_summary)

    if len(eval_part) > 0:
        print "[Evaluation part]"
        print ""
        eval_summary = get_summary(eval_part, ner, summary_groups)
        print_summary(eval_summary)



# encoding: utf-8

import os
from StringIO import StringIO
import tarfile

import model as crfmodel
from data.dataset import DataSet
from data.alphabet import Alphabet
from data import translate
from data.sentence import Sentence
from data.auxdata import AuxData
from train.sgd import sgd
from tager import Tager
from train.gradient import count_z_stats

#ALPHABET_FILE = "alphabet.gz"
#PARAMS_FILE = "params.gz"
#MISC_FILE = "misc.txt"

def train_crf(dataset, r=0.1, **kwargs):
    train, evl = dataset.divide_in_two(r)
    model = sgd(train, **kwargs)
    return CRF(model, dataset.alphabet, dataset.obtypes(),
            rev_alphabet=dataset.rev_alphabet)

class CRF:

    def __init__(self, model, alphabet, obtypes, rev_alphabet=None):
        self.model = model
        self.alphabet = alphabet
        self.rev_alphabet = rev_alphabet
        if self.rev_alphabet == None:
            self.rev_alphabet = self.alphabet.reversed()
        self.tager = Tager(model, obtypes)

    def save(self, filename):
        model_file = StringIO()
        self.model.save(fileobj=model_file)
        alph_file = StringIO()
        self.alphabet.save(fileobj=alph_file)

        tar = tarfile.open(filename, "w:gz")
        self._add_to_tar(tar, model_file, "params")
        self._add_to_tar(tar, alph_file, "alphabet")
        tar.close()

    @staticmethod
    def load(filename, obtypes):
        tar = tarfile.open(filename, "r:gz")
        modelfile = tar.extractfile("params")
        model = crfmodel.load(fileobj=modelfile)
        alphfile = tar.extractfile("alphabet")
        alphabet = Alphabet.load(fileobj=alphfile)
        crf = CRF(model, alphabet, obtypes)
        tar.close()
        return crf

    def tag(self, sent):
        """
        Tag sentence with a most probable sequence of labels, i.e. labels,
        which maximize P(labels | sent, CRF parameters) probability
        modeled by the conditional random field.
        """
        sent = Sentence(translate.sentence(sent, self.alphabet))
        labels = self.tager.tag(sent)
        return translate.labels(labels, self.rev_alphabet)

    def prob(self, sent, labels, log=True):
        """Compute P(labels | sent, CRF parameters)."""
        sent = Sentence(translate.sentence(sent, self.alphabet))
        labels = translate.labels(labels, self.alphabet)
        logz = count_z_stats([(sent, labels)], self.model,
                self.tager.connsbuf)[0]
        return self.model.prob(sent, labels, logz, log)

    def cll(self, rdata, regvar=4.0):
        """Conditional log likelihood of given data."""
        data = [(Sentence(translate.sentence(sent, self.alphabet)),
                translate.labels(labels, self.alphabet))
                for (sent, labels) in rdata]
        z_stats = count_z_stats(data, self.model, self.tager.connsbuf)
        i_var2 = 1.0 / (regvar ** 2)
        return self.model.cll(data, z_stats, 1.0, i_var2)

    def _add_to_tar(self, tar, fileobj, name):
        """Add to tar object fileobj with given name."""
        fileobj.seek(0)
        info = tarfile.TarInfo(name=name)
        info.size = len(fileobj.buf)
        tar.addfile(tarinfo=info, fileobj=fileobj)



# encoding: utf-8

cdef extern from "connsbuf.h":
    ctypedef struct C_ConnsBuf:
        pass
    C_ConnsBuf* newConnsBuf(int yn, int obstype_num)
    void freeConnsBuf(C_ConnsBuf* connsBuf)

cdef class ConnsBuf:
    cdef C_ConnsBuf* cbuf



# encoding=utf-8

import gzip

from crf import const

__all__ = ['Alphabet']

def sections(predicate, iterable):
    """sections(lambda x: x % 3 == 0, range(9))
    --> [[0, 1, 2], [3, 4, 5], [6, 7, 8]]"""
    value = []
    last = None
    for x in iterable:
        if predicate(x):
            if last != None:
                value.append(last)
            last = [x]
        else:
            last.append(x)
    value.append(last)
    return value

class Alphabet:
    """ Stores:
    - Observations alphabet,
    - Labels alphabet.
    """
	
    def __init__(self, void_label=None, dummy_label="DUMMY",
            observations=None, labels=None):
    	# Dummy label -- label for word before first word in sentence.
	self.dummy_label = dummy_label
	self.void_label = void_label

	self.observations = observations if observations != None else {}
        if labels != None:
            assert labels[self.dummy_label] == const.dummy_id
            self.labels = labels
        else:
            self.labels = {self.dummy_label: const.dummy_id}
            if self.void_label != None:
                self.add_label(self.void_label)

    def save(self, filename=None, fileobj=None):
        """Save alphabet in file_name using custom format."""
        if not fileobj:
            out = gzip.open(filename, 'wb')
        else:
            out = fileobj

        def fprint(*args):
            for arg in args:
                print >> out, arg.encode('utf-8'),
            print >> out, ""
        fprint("#DUMMY LABEL:", self.dummy_label)
        fprint("#VOID LABEL:", self.void_label)
        fprint("#OBSERVATIONS:")
        for k, v in self.observations.iteritems():
            fprint(k, str(v))
        fprint("#LABELS:")
        for k, v in self.labels.iteritems():
            fprint(k, str(v))

        if not fileobj:
            out.close()

    @staticmethod
    def load(filename=None, fileobj=None):
        """Load alphabet saved in custom format from filename."""

        def parse(inp):
            for section in sections(lambda line: line.startswith("#"), inp):
                name, value = map(lambda x: x.strip(),
                        section[0].lstrip("#").split(":"))
                yield name, value, section[1:]

        if not fileobj:
            inp = gzip.open(filename, 'rb')
        else:
            inp = fileobj

        header = {}
        observations = {}
        labels = {}
        for (name, value, contents) in parse(inp):
            if name in ["OBSERVATIONS", "LABELS"]:
                d = observations if name == "OBSERVATIONS" else labels
                for line in contents:
                    key, val = line.split()
                    d[key.decode('utf-8')] = int(val)
            else:
                header[name] = value
        return Alphabet(header["VOID LABEL"], header["DUMMY LABEL"],
                observations, labels)

    def observation(self, x):
	return self.observations.get(x)

    def label(self, x):
	return self.labels.get(x)

    def translate_feature(self, feat):
        (y1, y2, x) = feat
        if y1 != -1 and not self.labels.has_key(y1):
            return None
        if not self.labels.has_key(y2):
            return None
        if not self.observations.has_key(x):
            return None

        if y1 != -1:
            y1 = self.label(y1)
        y2 = self.label(y2)
        x = self.observation(x)
        return (y1, y2, x)
 
    def _add(self, alphabet, x):
	if not alphabet.has_key(x):
	    alphabet[x] = len(alphabet)

    def add_observation(self, x):
	self._add(self.observations, x)

    def add_label(self, x):
        if x == self.dummy_label:
            raise ValueError("label should be different than dummy label")
	self._add(self.labels, x)

    def reversed(self):
	if self.void_label != None:
	    void_label_tr = self.label(self.void_label)
	else:
	    void_label_tr = None
	rev_alph = Alphabet(void_label_tr, self.label(self.dummy_label))

	for label in self.labels:
	    rev_alph.labels[ self.label(label) ] = label

	for obs in self.observations:
	    rev_alph.observations[ self.observation(obs) ] = obs

	return rev_alph



#!/usr/bin/env python

import sys
import os

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from optparse import OptionParser
from ner.nerecognizer import NERecognizer

from tei.annotate import TEIAnnotator

if __name__ == "__main__":
    optparser = OptionParser(usage=
    """%prog CORPUS MODEL
	            
        Recognize NEs in CORPUS subdirectories using CRF model
        from MODEL file and write output to ann_named.xml.gz
        files in respective directories.""")

    optparser.add_option("-o", "--output_root", default=None,
            dest="output_root",
            help="Use another directory to store output.")
    optparser.add_option("-r", "--overwrite", action="store_true",
            default=False,
            dest="overwrite",
            help="Overwrite existing ann_named files.")
    optparser.add_option("--processes", type="int", default=1,
            dest="processes",
            help="number of processes in parallel annotation")

    (options, args) = optparser.parse_args()
    if len(args) != 2:
        optparser.print_help()
        sys.exit()

    corpus_path, model_file = args
    # annotator = TEIAnnotator(NERecognizer.load(model_file))
    annotator = TEIAnnotator(model_file, processes=options.processes)
    annotator.annotate(corpus_path,
            options.output_root,
            overwrite=options.overwrite)



import sys
import os
import traceback
import random
import re

from ner.features import sent2features, parse_feature_names
from ner.labels import encode, tuple2label

from read import TEIReader
#from atlas import atlas_nes

# TODO: Ponizsze przeniesc do skryptu nes_from_corpus_to_corpus.
#def flatten(l):
#    v = []
#    for x in l:
#        v.extend(x)
#    return v
#
def corpus_dirs(root, ann_named="ann_named.xml"):
    for desc_dir, dirs, files in os.walk(root):
        if (('ann_morphosyntax.xml' in files or
            'ann_morphosyntax.xml.gz' in files) and
            ann_named in files):
            yield re.sub(root, "", desc_dir).lstrip("/")

def parse_dir(dirname, features, ann_named="ann_named.xml"):
    reader = TEIReader(dirname)
    names = list(reader.names(ann_named))
    names, segs = zip(*names)
    data = []
    for (sent_names, sent_segs) in zip(names, segs):
        words = sent2features(sent_segs, features)
        labels = map(tuple2label, encode(sent_names, sent_segs))
        data.append((
            (sent_segs, sent_names), (words, labels)
            ))
    return data

def read_tei_corpus(tei_corpus, schema, black_listed=[], select=1.0,
        ann_named="ann_named.xml", verbose=False):
    """
    Read data from TEI corpus.
    
    :params:
    --------
    tei_corpus : path
        Path to TEI corpus.
    schema : list
        List of observation types included.
    black_listed : list
        Skip corpus subdirectories present in the list.
    select : float in [0.0, 1.0]
        Select random $select * 100% files from corpus.
    ann_named : str
        Use this name to find NE files.
    """
    features = parse_feature_names(schema)
    raw_data, dirs = [], []
    random.seed(0)
    for dirname in corpus_dirs(tei_corpus, ann_named):
        if (any(dirname.endswith(black) for black in black_listed) or
                random.random() > select):
            if verbose:
                print "Skipping directory", dirname
            continue
        elif verbose:
            print "Reading directory", dirname
        try:
            sent_data = parse_dir(os.path.join(tei_corpus, dirname),
                    features, ann_named)
            raw_data.extend(sent_data)
            dirs.extend([dirname] * len(sent_data))
        except:
            info = sys.exc_info()
            print "Exception:", info[1]
            traceback.print_tb(info[2], file=sys.stdout)
    return raw_data, dirs




#def flatten(l):
#    v = []
#    for elem in l:
#        v.extend(elem)
#    return v

def flatten_pairs(l):
    for (l1, l2) in l:
        for x, y in zip(l1, l2):
            yield (x, y)

def _stats(pairs, void_label):
    """Return precision, recall, F-measure."""
    tp = 0 # true positives
    fp = 0 # false positives
    tn = 0 # true negatives
    fn = 0 # false negatives

    for x, y in pairs:
        if x == void_label:
            if y == void_label:
                tn = tn + 1
            else:
                fp = fp + 1
        if x != void_label:
            if y == x:
                tp = tp + 1
            else:
                fn = fn + 1
                if y != void_label:
                    fp = fp + 1
    
    prec = rec = 0.0
    if tp + fp != 0: prec = tp / float(tp + fp)
    if tp + fn != 0: rec = tp / float(tp + fn)
    if prec == 0.0 and rec == 0.0:
    	return (0.0, 0.0, 0.0)
    F = (2*prec*rec) / (prec + rec)
    return (prec, rec, F)

def stats(data, tager, void_id):
    """Compute statistics scored with tager on given data."""
    labels_and_tagged = ((labels, tager.tag(sent)) for (sent, labels) in data)
    return _stats(flatten_pairs(labels_and_tagged), void_id)



# encoding: utf-8

from train.sgd import sgd
from tager import Tager
from stats import stats
from model import Model
from data.dataset import features_in_data

def flatten(l):
    v = []
    for x in l:
        v.extend(x)
    return v

def cross_validation(dataset, k, *sgd_args, **sgd_kwargs):
    """Perform k-cross validation on the given dataset."""
    train_prec = []
    train_rec = []
    train_F = []
    eval_prec = []
    eval_rec = []
    eval_F = []

    void_id = dataset.alphabet.label(unicode(dataset.alphabet.void_label))
    parts = dataset.divide(k)
    obtypes = dataset.obtypes()
    for i in range(k):
        print "======== Part %s ========" % i
        print ""

        evl = parts[i]
        train = flatten(parts[:i] + parts[i+1:])
        model = Model(features_in_data(train))
        sgd(model, train, *sgd_args, **sgd_kwargs)
        tager = Tager(model, obtypes)

        _prec, _rec, _F = stats(train, tager, void_id)
        print "Train part:"
        print "Precision:", _prec
        print "Recall:", _rec
        print "F-measure:", _F
        print ""
        train_prec.append(_prec)
        train_rec.append(_rec)
        train_F.append(_F)

        _prec, _rec, _F = stats(evl, tager, void_id)
        print "Eval part:"
        print "Precision:", _prec
        print "Recall:", _rec
        print "F-measure:", _F
        print ""
        eval_prec.append(_prec)
        eval_rec.append(_rec)
        eval_F.append(_F)

    train_stats = (train_prec, train_rec, train_F)
    eval_stats = (eval_prec, eval_rec, eval_F)
    return train_stats, eval_stats



# encoding: utf-8

cdef class Conns:

    def __cinit__(self):
        self.conns.xn = 0
        self.conns.conns = NULL
        self.conns.csizes = NULL 

    def _dealloc(self):
        if self.conns.csizes is not NULL:
            free(self.conns.csizes)
        if self.conns.conns is not NULL:
            for i in range(self.conns.xn):
                if self.conns.conns[i] is not NULL:
                    free(self.conns.conns[i])
            free(self.conns.conns)

    def __dealloc__(self):
        self._dealloc()

    def __init__(self, Model model):
        self._dealloc()
        self.conns.xn = model.xn

        conns = defaultdict(list)
        for i, (feature, _) in enumerate(model):
            (py, y, x) = feature
            conns[x].append((py, y, i))
        # conns = sorted(conns.items())
        # assert len(conns) == self.conns.xn

        self.conns.csizes = <int*>malloc(self.conns.xn * sizeof(int))
        if self.conns.csizes is NULL:
            raise MemoryError
        self.conns.conns = <int**>malloc(self.conns.xn * sizeof(int*))
        if self.conns.conns is NULL:
            raise MemoryError
        for x in range(self.conns.xn):
            self.conns.conns[x] = NULL

        for x in range(self.conns.xn):
            xconns = conns.get(x, [])
            self.conns.csizes[x] = len(xconns)
            self.conns.conns[x] = <int*>malloc(
                    max(1, len(xconns)) * 3 * sizeof(int))
            if self.conns.conns[x] is NULL:
                raise MemoryError
            for k, (py, y, par_id) in enumerate(xconns):
                self.conns.conns[x][k*3] = py
                self.conns.conns[x][k*3 + 1] = y
                self.conns.conns[x][k*3 + 2] = par_id

    def __getitem__(self, x):
        if x < 0 or x >= self.conns.xn:
            raise IndexError
        return [(
            self.conns.conns[x][k*3],
            self.conns.conns[x][k*3 + 1],
            self.conns.conns[x][k*3 + 2])
            for k in range(self.conns.csizes[x])]



# encoding: utf-8

import os
import tarfile
from StringIO import StringIO

from crf.api import CRF
import crf.model as crfmodel
from crf.data.alphabet import Alphabet

from data import DummySegment
from features import sent2features, parse_feature_names
from labels import decode, label2tuple

class NERecognizer(CRF):

    def __init__(self, model, alphabet, schema, rev_alphabet=None):
        self.schema = schema
        self.features = parse_feature_names(schema)
        # print "number of features:", len(self.features)
        CRF.__init__(self, model, alphabet, len(self.features),
                rev_alphabet=rev_alphabet)

    def save(self, filename):
        model_file = StringIO()
        self.model.save(fileobj=model_file)
        alph_file = StringIO()
        self.alphabet.save(fileobj=alph_file)

        tar = tarfile.open(filename, "w:gz")
        self._add_to_tar(tar, model_file, "params")
        self._add_to_tar(tar, alph_file, "alphabet")
        self._add_to_tar(tar, StringIO("\n".join(self.schema)), "schema")
        tar.close()

    @staticmethod
    def load(filename):
        tar = tarfile.open(filename, "r:gz")
        modelfile = tar.extractfile("params")
        model = crfmodel.load(fileobj=modelfile)
        alphfile = tar.extractfile("alphabet")
        alphabet = Alphabet.load(fileobj=alphfile)
        featfile = tar.extractfile("schema")
        schema = map(lambda x: x.strip(), featfile.readlines())
        ner = NERecognizer(model, alphabet, schema)
        tar.close()
        return ner

    def prob(self, sent, labels, log=True):
        raise NotImplementedError
#        """Compute P(labels | sent, CRF parameters)."""
#        sent = Sentence(translate.sentence(sent, self.alphabet))
#        labels = translate.labels(labels, self.alphabet)
#        logz = count_z_stats([(sent, labels)], self.model,
#                self.tager.connsbuf)[0]
#        return self.model.prob(sent, labels, logz, log)

    def cll(self, rdata, regvar=4.0):
        raise NotImplementedError
#        """Conditional log likelihood of given data."""
#        data = [(Sentence(translate.sentence(sent, self.alphabet)),
#                translate.labels(labels, self.alphabet))
#                for (sent, labels) in rdata]
#        z_stats = count_z_stats(data, self.model, self.tager.connsbuf)
#        i_var2 = 1.0 / (regvar ** 2)
#        return self.model.cll(data, z_stats, 1.0, i_var2)

    def recognize_named_entities(self, sent, features=None):
        """
        Recognize NEs in the sentence.

        :param features:
            Feature values already extracted from sentence.
        """
        if not features:
            features = sent2features(sent, self.features)
        labels = [label2tuple(y) for y in self.tag(features)]
        return decode(labels, sent, recover_not_started=True, max_dist=2)



#!/usr/bin/env python

import sys
import os
from collections import defaultdict

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

# from crf.data.dataset import DataSet, divide_in_two
from ner.data import NerDataSet
from ner.nerecognizer import NERecognizer
from tei.stats import count_names, count_errors
from tei.corpus import read_tei_corpus

def file_lines(filename):
    return map(lambda x: x.strip(), open(filename).readlines())

def netype(s):
    return s.split(":")[0]

def group_types(stats, by_type=False):
    result = defaultdict(int)
    for key, val in stats.iteritems():
        if by_type:
            key = netype(key)
        result[key] += val
    return result

def group_pairs(stats, by_type=False):
    result = defaultdict(int)
    for (key1, key2), val in stats.iteritems():
        if by_type:
            key1 = netype(key1)
            key2 = netype(key2)
        result[(key1, key2)] += val
    return result

def print_stats(stats):
    for key, val in reversed(sorted(stats.items(), key=lambda (k, v): v)):
        print key, "=>", val

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog,
        
        Extract data from tei_corpus, divide it to training part and
        evaluation part and perform error analisys on the evaluation
        part.""")
	
    optparser.add_option("--save-errors", metavar="FILE", dest="save_errors",
            help="Save list of errors in.")
    optparser.add_option("--tei-corpus", metavar="DIR", dest="tei_corpus",
            help="Use TEI corpus as data source.")
    optparser.add_option("--select", type="float", default=1.0,
            dest="select", help="select subcorpus of a given relative size.")
    optparser.add_option("--relative-eval-size", type="float",
            default=0.1, dest="eval_size",
            help="Relative (to the data size) size of the eval part.")
    optparser.add_option("--model-in", metavar="FILE", dest="model_in",
            help="start training from initial model")
    optparser.add_option("--ann-named", type="string",
            dest="ann_named", default="ann_named.xml",
            help="Names of input ann_named files.")
    optparser.add_option("--black-listed", metavar="FILE",
            dest="black_listed", help="black listed TEI corpus directories")
#    optparser.add_option("--void-label", type="str", dest="void_label",
#            default="O", help="label representing void information")
#    optparser.add_option("--dummy-label", type="str", dest="dummy_label",
#            default="DUMMY", help="label corresponding to the posision in"
#            " a sentence before the first word")

    (options, args) = optparser.parse_args()
    if options.tei_corpus == None:
        optparser.print_help()
        print "\n--tei-corpus option is mandatory"
        sys.exit(0)
    if options.model_in == None:
        optparser.print_help()
        print "\n--model-in option should be supplied"
        sys.exit(0)

    print "Loading model from %s..." % options.model_in
    ner = NERecognizer.load(options.model_in)

    if options.black_listed != None:
        black_listed = file_lines(options.black_listed)
    else:
        black_listed = []

    print "Reading data..."
    raw_data, dirs = read_tei_corpus(options.tei_corpus, ner.schema,
            black_listed, options.select, ann_named=options.ann_named)
    dataset = NerDataSet(raw_data, alphabet=ner.alphabet, dirs=dirs)
    print "Reading done\n"

    train_part, eval_part = dataset.divide_in_two(options.eval_size)
    print "Evaluation part statistics:"
    print "Number of sentences:", len(eval_part)
    print "Number of segments:", sum(len(sent)
            for sent, labels in eval_part)
    print "Number of NEs:", count_names(eval_part)
    print ""

    tp, fp, fn, substs, pairs, fn_errors, fp_errors,\
            subst_errors, pair_errors = count_errors(eval_part, ner)
    print "=== TRUE POSITIVES ==="
    print_stats(group_types(tp))
    print "=== FALSE POSITIVES ==="
    print_stats(group_types(fp))
    print "=== FALSE NEGATIVES ==="
    print_stats(group_types(fn))
    print "=== SUBSTITUTIONS ==="
    print_stats(group_pairs(substs))
    print "=== PAIRS ==="
    print_stats(group_pairs(pairs))

    if options.save_errors is not None:
        with open(options.save_errors, "w") as f:
            print >> f, "FALSE NEGATIVE ERRORS:"
            for (t, s), sent in fn_errors:
                print >> f, "#", sent
                print >> f, t, s.encode('utf-8'), "=>"
            print >> f, "FALSE POSITIVE ERRORS:"
            for (t, s), sent in fp_errors:
                print >> f, "#", sent
                print >> f, "=>", t, s.encode('utf-8')
            print >> f, "SUBSTITUTION ERRORS:"
            for (t1, s1), (t2, s2), sent in subst_errors:
                print >> f, "#", sent
                print >> f, t1, s1.encode('utf-8'), "=>", t2, s2.encode('utf-8')
            print >> f, "PAIR ERRORS:"
            for (t1, s1), (t2, s2), sent in pair_errors:
                print >> f, "#", sent
                print >> f, t1, s1.encode('utf-8'), "=>", t2, s2.encode('utf-8')



#!/usr/bin/env python

import sys
import os

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

from crf.model import Model
from crf.data.dataset import features_in_data
from crf.train.sgd import sgd

from ner.nerecognizer import NERecognizer
from ner.data import NerDataSet, filter_raw

from tei.corpus import read_tei_corpus
from tei.stats import name_stats, count_names, old_count_names
from train import get_summary, get_overall, print_summary, nkjp_summary_groups
from train import CallBack

def flatten(l):
    result = []
    for subl in l:
        result.extend(subl)
    return result

def file_lines(filename):
    return map(lambda x: x.strip(), open(filename).readlines())

def avg(l):
    return float(sum(l)) / len(l)

def cross_validation(dataset, schema, k, save_models_in=None,
        pre_train=False, sgd_kwargs={}, eval_every=10.0):
    """Perform k-cross validation on the given dataset."""
    if save_models_in is not None:
        if not os.path.isdir(save_models_in):
            os.mkdir(save_models_in)

    void_id = dataset.alphabet.label(unicode(dataset.alphabet.void_label))
    parts = dataset.divide(k,
            # stat=lambda x: old_count_names([x], dataset.rev_alphabet))
            stat=lambda x: count_names([x]))

    print "Data statistics:"
    for i in range(k):
        print "Part %s:" % i
        print "Number of sentences:", len(parts[i])
        # print "Number of NEs:", count_names(parts[i], dataset.rev_alphabet)
        print "Number of NEs:", count_names(parts[i])
    print ""

    summaries = []
    for i in range(k):
        print "======== Part %s =========" % i
        print ""

        evl = parts[i]
        train = flatten(parts[:i] + parts[i+1:])
        # if i > 2:
        if True:
            model = Model(features_in_data(filter_raw(train)))
            ner = NERecognizer(model, dataset.alphabet, schema)
            callback = CallBack(train, evl, ner, every=eval_every)

            if pre_train:            
                print "Pre-training model..."
                sgd(model, filter_raw(train), callback=callback, **{
                    "iter_num": 10,
                    "batch_size": sgd_kwargs["batch_size"],
                    "tau": sgd_kwargs["tau"],
                    "scale0": 0.01,
                    "workers": sgd_kwargs["workers"],
                    "verbose": sgd_kwargs["verbose"],
                    })
                print "Pre-training done\n"

            print "Training model..."
            sgd(model, filter_raw(train), callback=callback, **sgd_kwargs)
            print "Training done\n"

            if save_models_in is not None:
                save_path = os.path.join(save_models_in, "model%s.tgz" % i)
                print "\nSaving model in %s..." % save_path
                ner.save(save_path)
                print "Done"
        else:
            assert save_models_in is not None
            path = os.path.join(save_models_in, "model%s.tgz" % i)
            print "Loading model from %s..." % path
            ner = NERecognizer.load(path)
            print "Done"

        print ""
        print "==== Summary for part %s ====" % i
        print ""
        print "[Train part]"
        print ""
        train_summary = get_overall(train, ner)
        print_summary(train_summary)

        print "[Evaluation part]"
        print ""
        eval_summary = get_summary(evl, ner, nkjp_summary_groups())
        print_summary(eval_summary)

        summaries.append((train_summary, eval_summary))
    return summaries

def averaged_summary(summaries):
    results = {}
    cats = set()
    for summary in summaries:
        for category, stats in summary.iteritems():
            cats.add(category)
            if not results.has_key(category):
                results[category] = []
            results[category].append(stats)
    averaged = {}
    for category in cats:
        # lista wszytkich statystyk (prec, recall, ...) na kolejnych czesciach
        stat_list = results[category]

        types = ["Precision", "Recall", "F-measure"]
        cat_averaged = []
        for type in types:
            val = avg([dict(stats)[type] for stats in stat_list])
            cat_averaged.append((type, val))

        averaged[category] = tuple(cat_averaged)
    return averaged

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog,
        
        Extract data from corpus, divide it to k parts and
        perform k-cross validation.""")
	
    optparser.add_option("-k", "--k-cross-validation", type="int", default=10,
            dest="k", help="Perform k-cross validation.")
    optparser.add_option("--tei-corpus", metavar="DIR", dest="tei_corpus",
            help="Use TEI corpus as data source.")
    optparser.add_option("--ann-named", type="string",
            dest="ann_named", default="ann_named.xml",
            help="Names of input ann_named files.")
    optparser.add_option("--data-out", metavar="FILE",
            dest="data_out", help="Data output file (for debugging purpose).")
    optparser.add_option("--select", type="float", default=1.0,
            dest="select", help="Select subcorpus of given relative size.")
    optparser.add_option("--schema", metavar="FILE",
            dest="schema", help="File with feature schema.")
    optparser.add_option("--black-listed", metavar="FILE",
            dest="black_listed", help="Black listed corpus directories.")
    optparser.add_option("--batch-size", type="int",
            default=30, dest="batch_size",
            help="Batch size in SGD method.")
    optparser.add_option("--evaluate-every", type="int",
            default=10, dest="eval_every",
            help="Compute stats on evaluation part every * interations.")
    optparser.add_option("--iter-num", type="int", default=100,
            dest="iter_num", help="Number of SGD method iterations.")
    optparser.add_option("--threads", type="int", default=1, dest="threads",
            help="Number of threads in parallel computation.")
    optparser.add_option("--regvar", type="float", dest="regvar",
            help="Regularization variance parameter, see SGD specification.")
    optparser.add_option("--scale0", type="float", default=0.01,
            dest="scale0", help="Scale0 parameter, see SGD specification.")
    optparser.add_option("--tau", type="float", default=10.0,
            dest="tau", help="Tau parameter, see SGD specification.")
    optparser.add_option("--pre-train", default=False,
            action="store_true", dest="pre_train",
            help="Pre-train model on input data with predefined," +
                    " small scale0 value.")
    optparser.add_option("--void-label", type="str", dest="void_label",
            default="O", help="Label representing void information.")
    optparser.add_option("--dummy-label", type="str", dest="dummy_label",
            default="DUMMY", help="Label corresponding to the posision in"
            " a sentence before the first word.")
    optparser.add_option("--save-models-in", metavar="DIR",
            dest="save_models_in",
            help="Save models created during the process in given directory.")

    (options, args) = optparser.parse_args()
    if options.tei_corpus == None:
        optparser.print_help()
        print "\n--tei_corpus option is mandatory"
        sys.exit(0)
    if options.schema == None:
        optparser.print_help()
        print "\n--schema option should be supplied"
        sys.exit(0)

    if options.black_listed != None:
        black_listed = file_lines(options.black_listed)
    else:
        black_listed = []

    print "Reading data..."
    schema = file_lines(options.schema)
    raw_data, dirs = read_tei_corpus(options.tei_corpus, schema,
            black_listed, options.select, ann_named=options.ann_named)
    dataset = NerDataSet(raw_data, void_label=unicode(options.void_label),
            dummy_label=unicode(options.dummy_label), dirs=dirs)
    print "Reading done\n"

    if options.data_out:
        dataset.print_data(options.data_out)

    summaries = cross_validation(dataset, schema, options.k,
            save_models_in=options.save_models_in,
            pre_train=options.pre_train,
            eval_every=options.eval_every,
            sgd_kwargs={
                "iter_num": options.iter_num,
                "batch_size": options.batch_size,
                "tau": options.tau,
                "scale0": options.scale0,
                "workers": options.threads,
                "verbose": False
                })

    print "====== On average ======"
    print ""
    print "[Train part]"
    print ""
    print_summary(averaged_summary(train_summ for train_summ, _ in summaries))

    print "[Evaluation part]"
    print ""
    print_summary(averaged_summary(eval_summ for _, eval_summ in summaries))



'''
Created on 17-03-2011

@author: lennyn
'''
import xml.sax
from ner.labels import dicts2names
from ner.data import NamedEntity

class NamesHandler(xml.sax.ContentHandler):
    '''
    classdocs
    '''


    def __init__(self, queue):
        '''
        Constructor
        '''
        self.queue = queue
        self.curr_names = None
        
        self.fname = None
        self.in_string = False
        
        self.ne_id = None
        self.type = None
        self.subtype = None
        self.derivType = None
        self.derivedFrom = None
        self.ptrs = None
        
        self.id2interp = None
    
    def startElement(self, name, attrs):
        if name == 's':
            self.curr_names = []
        elif name == 'seg':
            self.ne_id = attrs.getValue('xml:id')
            self.type = None
            self.subtype = None
            self.derivedFrom = None
            self.derivType = None
            self.ptrs = []
        elif name == 'f':
            self.fname = attrs.getValue('name')
        elif name == 'string':
            self.in_string = True
        elif name == 'symbol':
            if self.fname == 'derivType':
                self.derivType = attrs.getValue('value')
            elif self.fname == 'type':
                self.type = attrs.getValue('value')
            elif self.fname == 'subtype':
                self.subtype = attrs.getValue('value')
        elif name == 'ptr':
            self.ptrs.append(attrs.getValue('target').split('#')[-1])
    
    def endElement(self, name):
        if name == 's':
            self.queue.put(self.curr_names)
        elif name == 'seg':
            self.curr_names.append({
                            'id' : self.ne_id,
                            'type' : self.type,
                            'subtype' : self.subtype,
                            'ptrs' : self.ptrs,
                            'derivType' : self.derivType,
                            'derivedFrom' : self.derivedFrom,
                            })
    
    def characters(self, content):
        if self.in_string:
            if self.fname == 'derivedFrom':
                self.derivedFrom = content.strip()
    
    def endDocument(self):
        self.queue.put(None)



import glob
from distutils.core import setup
from distutils.extension import Extension

from Cython.Distutils import build_ext

ext_modules = [
        Extension("crf.model", ["crf/model.pyx", "crf/conns.pxi"],
            include_dirs=["crf/libcrf"],
            library_dirs=["crf/libcrf"], libraries=["fastcrf"]),
        Extension("crf.array", ["crf/array.pyx"]),
        Extension("crf.data.sentence", ["crf/data/sentence.pyx"]),
        Extension("crf.data.connsbuf", ["crf/data/connsbuf.pyx"],
            include_dirs=["crf/libcrf"],
            library_dirs=["crf/libcrf"], libraries=["fastcrf"]),
        Extension("crf.data.auxdata", ["crf/data/auxdata.pyx"]),
        Extension("crf.train.sgd", ["crf/train/sgd.pyx"],
            include_dirs=["crf/libcrf"]),
        Extension("crf.train.gradient", ["crf/train/gradient.pyx"],
            include_dirs=["crf/libcrf"],
            library_dirs=["crf/libcrf"], libraries=["fastcrf"]),
        Extension("crf.tager", ["crf/tager.pyx"],
            include_dirs=["crf/libcrf"],
            library_dirs=["crf/libcrf"], libraries=["fastcrf"]),
    ]

setup(name="pycrf",
        version="0.2",
        author='Jakub Waszczuk',
        author_email='waszczuk.kuba@gmail.com',
        cmdclass={'build_ext': build_ext},
        ext_modules=ext_modules,
        packages=["crf", "crf.data", "crf.train"])



# encoding: utf-8

import sys
import os

# Solution with no hard coded path would be welcome
sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + "/../")

import traceback
from itertools import izip
import copy
from collections import defaultdict

from ner.data import Segment, NamedEntity

from tei.read import TEIReader
import tei.write as write
from tei.corpus import corpus_dirs

# Czy rozwiazywac ewentualne roznice pomiedzy segmentacjami ?
REPAIR = True

def flatten(l):
    result = []
    for subl in l:
        result.extend(subl)
    return result

def LCS(X, Y):
    m = len(X)
    n = len(Y)
    # An (m+1) times (n+1) matrix
    C = [[0] * (n+1) for i in range(m+1)]
    for i in range(1, m+1):
        for j in range(1, n+1):
            if X[i-1] == Y[j-1]: 
                C[i][j] = C[i-1][j-1] + 1
            else:
                C[i][j] = max(C[i][j-1], C[i-1][j])
    return C[m][n]

def sameness(s1, s2):
    lcs = float(LCS(s1, s2))
    return min(lcs / len(s1), lcs / len(s2))

def orth(x):
    return x.orth if x.nps else " " + x.orth

def getfirst(l, k):
    res = ""
    for x in l:
        res += orth(x)
        if len(res) >= k:
            break
    return res[:k]

def getlast(l, k):
    res = ""
    for x in reversed(l):
        res = orth(x) + res
        if len(res) >= k:
            break
    return res[-k:]

def signal_diff(main, other, main_done, other_done, m, n, window_width=30):
    ww = window_width
    if REPAIR and sameness(getfirst(main, ww), getfirst(other, ww)) > 0.75:
        pass
    else:
        k = m - n
        main_orth = getlast(main_done, ww) + getfirst(main, ww)
        other_orth = getlast(other_done, ww - k) + getfirst(other, ww + k)
        print "FILE1:", main_orth.encode('utf-8'), "["+str(main[0].id)+"]"
        print "FILE2:", other_orth.encode('utf-8'), "["+str(other[0].id)+"]"
        k = 6 if main[0].nps else 7
        print " " * (k + len(getlast(main_done, ww))), "▲"
        # print "[" + str(sameness(main_orth, other_orth)) + "]"
        raise Exception("synchronization error")

def orth_eq(orth1, orth2):
    orth1 = orth1.replace(u"­", "-").replace(u"–", "-")
    orth2 = orth2.replace(u"­", "-").replace(u"–", "-")
    return orth1 == orth2

def align_map(main, other):
    """
    Align two sequences of segments.

    Return map (seg \in main) -> (segs \subset other).
    """
    main = copy.copy(main)
    other = copy.copy(other)
    main_done = []
    other_done = []
    m, n = 0, 0
    map = defaultdict(set)
    # log = [] #@
    while len(main) > 0 and len(other) > 0:
        main_orth = main[0].orth
        # main_orth = orth(main[0])
        other_orth = other[0].orth
        # other_orth = orth(other[0])
        # log.append((main_orth, other_orth)) #@
        map[main[0]].add(other[0])
        imin, imax = max(m, n), min(m + len(main_orth), n + len(other_orth))
        if not orth_eq(main_orth[imin - m : imax - m],
                other_orth[imin - n : imax - n]):
            # import pdb; pdb.set_trace()
            signal_diff(main, other, main_done, other_done, m, n)
        if m + len(main_orth) == n + len(other_orth):
            m += len(main_orth); main_done.append(main.pop(0))
            n += len(other_orth); other_done.append(other.pop(0))
        elif m + len(main_orth) < n + len(other_orth):
            m += len(main_orth); main_done.append(main.pop(0))
        else:
            n += len(other_orth); other_done.append(other.pop(0))
    if max(len(main), len(other)) > 0:
        # import pdb; pdb.set_trace()
        raise Exception("synchronization failed")
    return map

def substitute_segs(name, map):
    """Substitute map[segment] for each segment descendant."""
    segs = name.get_child_segs() 
    segs = flatten(map[seg] for seg in segs)
    nes = name.get_child_names()
    nes = [substitute_segs(ne, map) for ne in nes]
    return NamedEntity(name.id,
            name.type,
            name.subtype, 
            ptrs=(nes + segs),
            base=name.base,
            derivType=name.derivType,
            derivedFrom=name.derivedFrom)

#def repair_name(name):
#    """Repair NE if corrupted after segmentation synchronization."""
#    nes_with_segs = [repair_name(ne) for ne in name.get_child_names()]
#    segs = name.get_child_segs()
#    seg_sets = dict((ne, set(ne.get_segs())) for ne in nes)
#
#    # remove NE children when their segment sets intersect
#    nes_ok = []
#    def check(ne, nes, seg_sets):
#        for other_ne in nes:
#            if len(seg_sets[ne] & seg_sets[other_ne]) > 0:
#                nes.remove(other_ne)
#                return False
#        return True
#    while len(nes) > 0:
#        ne = nes.pop()
#        if check(ne, nes, seg_sets):
#            nes_ok.append(ne)
#
#    return NamedEntity(name.id,
#            name.type,
#            name.subtype, 
#            ptrs=(nes_ok + segs),
#            base=name.base,
#            derivType=name.derivType,
#            derivedFrom=name.derivedFrom)

def _add_to_group(name, names_grouped, segmap):
    name_segs = name.get_segs()
    name_segs = set(segmap[seg] for seg in name_segs)
    if len(name_segs) > 1:
        print "WARNING: Named Entity on intersection of sentences"
        print name
        for ne in name.get_child_names():
            _add_to_group(ne, names_grouped, segmap)
    else:
        names_grouped[name_segs.pop()].append(name)

def group_names(names, segs):
    # segment -> sentence map
    segmap = {}
    names_grouped = {}
    for i, sent_seg in enumerate(segs):
        names_grouped[i] = []
        for seg in sent_seg:
            segmap[seg] = i
    for name in names:
        _add_to_group(name, names_grouped, segmap)
    return [sent_names for i, sent_names in sorted(names_grouped.items())]

#def sync_with_segs(fromnames, fromsent, tosent):
#    map = align_map(fromsent, tosent)
#    names = [substitute_segs(name, map)
#            for name in fromnames]
#    # names = [repair_name(name) for name in names]
#    # return group_names(names, morph_segs), morph_segs
#    return names

def sync_with_segs(fromnames, fromsegs, tosegs):
    map = align_map(flatten(fromsegs), flatten(tosegs))
    names = [substitute_segs(name, map)
            for sent_names in fromnames
            for name in sent_names]
    # names = [repair_name(name) for name in names]
    return group_names(names, tosegs)

#def sync_contents(from_contents, to_contents):
#    fromids, fromsents, fromnames = zip(*from_contents)
#    toids, tosents = zip(*to_contents)
#    names = sync_with_segs(fromnames, fromsents, tosents)
#    for sent_id, sent_names in zip(toids, names):
#        yield sent_id, sent_names
#
#    for (fromid, fromsent, fromnames), (toid, tosent) in izip_longest(
#            from_contents, to_contents):
#        names = sync_with_segs(fromnames, fromsent, tosent)
#        yield from, names


def sync_names(fromdir, todir, ann_named="ann_named.xml"):
    fromreader = TEIReader(fromdir)
    toreader = TEIReader(todir)

    fromnames, fromsegs = zip(*fromreader.names(ann_named))
    tosegs = list(toreader.segments())
    
    names = sync_with_segs(fromnames, fromsegs, tosegs)

    for parid, sents in toreader.paragraphs():
        snum = len(list(sents))
        yield names[:snum]
        names = names[snum:]

    assert len(names) == 0

#    for (from_parid, from_contents), (to_parid, to_contents) in izip_longest(
#            from_reader.paragraphs_with_names(ann_named),
#            to_reader.paragraphs()):
#        yield (to_parid, sync_contents(from_contents, to_contents))

def iter_contents(parcontents, parnames):
    for names, (sentid, contents) in izip(parnames, parcontents):
        yield sentid, names

def converted(fromdir, todir, ann_named="ann_named.xml"):
    names = sync_names(fromdir, todir, ann_named)
    reader = TEIReader(todir)
    for parnames, (parid, contents) in izip(names, reader.paragraphs()):
        yield (parid, iter_contents(contents, parnames))

def move_names(from_tei, to_tei, ann_named="ann_named.xml"):
    """
    Move ann_named.xml files from TEI corpus to another TEI corpus
    (with possibly different annotations on lower levels).
    
    :params:
    --------
    from_tei : path
        Path to TEI corpus.
    to_tei : path
        Path to another TEI corpus.
    ann_named : str
        Use this name to find NE files.
    """
    for dirname in corpus_dirs(from_tei, ann_named):
        print "Directory:", dirname
        try:
            fromdir = os.path.join(from_tei, dirname)
            todir = os.path.join(to_tei, dirname)
            with open(os.path.join(todir, ann_named), "w") as out:
                write.write(out, converted(fromdir, todir, ann_named))
        except:
            print traceback.format_exc()

if __name__ == "__main__":
    from optparse import OptionParser
    optparser = OptionParser(usage="""usage: %prog,

        Move ann_named.xml files from TEI corpus to another TEI corpus
        (with, possibly, different annotations on lower levels).""")
	
    optparser.add_option("--from-tei", metavar="DIR", dest="from_tei")
    optparser.add_option("--to-tei", metavar="DIR", dest="to_tei")
    optparser.add_option("--ann-named", type="string",
            dest="ann_named", default="ann_named.xml",
            help="Names of ann_named files.")

    (options, args) = optparser.parse_args()
    if options.from_tei == None:
        optparser.print_help()
        print "\n--from-tei option is mandatory"
        sys.exit(0)
    if options.to_tei == None:
        optparser.print_help()
        print "\n--to-tei option is mandatory"
        sys.exit(0)

    move_names(options.from_tei, options.to_tei, options.ann_named)



import os
import re
import gzip
from itertools import takewhile, repeat, starmap, imap, groupby, izip_longest
from itertools import islice
from Queue import Queue
from threading import Thread

import xml.sax

from morph_handler import MorphHandler
from names_handler import NamesHandler
from ner.labels import dicts2names

MAX_Q_SIZE = 1

def repeatfunc(func, times=None, *args):
    """Repeat calls to func with specified arguments.

    Example:  repeatfunc(random.random)
    """
    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))

def queue_iterator(queue):
    return takewhile(lambda x: x != None, repeatfunc(queue.get))

def take(n, iterable):
    "Return first n items of the iterable as a list"
    return list(islice(iterable, n))

class TEIReader(object):
    '''
    classdocs
    '''

    def __init__(self, dirname):
        self._dirname = dirname
        morph_gz = os.path.join(dirname, 'ann_morphosyntax.xml.gz')
        self._gzipped = os.path.isfile(morph_gz)

    def ids(self):
        """Return paragraph id and sentence ids for each paragraph."""
	# par_start = re.compile("\s*<%s [^>]*>\s*" % "p")
	par_start = re.compile("<p ")
	par_end = re.compile("</p>")
	sent_start = re.compile("<s ")
	sent_end = re.compile("</s>")
        get_id = re.compile("xml:id=\"(.*?)\"")

        par_id = None
        sent_ids = None

        if self._gzipped:
            f = gzip.open(os.path.join(self._dirname, 'ann_morphosyntax.xml.gz'))
        else:
            f = open(os.path.join(self._dirname, 'ann_morphosyntax.xml'))

        for line in f:
	    if par_start.search(line):
                par_id = get_id.search(line).group(1)
                sent_ids = []
            elif par_end.search(line):
                yield par_id, sent_ids
	    elif sent_start.search(line):
                sent_ids.append(get_id.search(line).group(1))
        f.close()

    def segments(self):
        """Return segments for each sentence."""
        queue = Queue(maxsize=MAX_Q_SIZE)
        thread = Thread(target=self._parse_morph, args=(queue,))
        thread.start()
        for par_id, sent_id, sent_segs in queue_iterator(queue):
            yield sent_segs
        thread.join()

    def names(self, ann_named):
        """Return names for each sentence."""
        segments = self.segments()
        queue = Queue(maxsize=MAX_Q_SIZE)
        thread = Thread(target=self._parse_names, args=(ann_named, queue))
        thread.start()
        for sent_names, sent_segs in izip_longest(queue_iterator(queue), segments):
            # Probably different size of queues and different number of
            # sentences in ann_morph/ann_named files.
            assert sent_names != None and sent_segs != None
            yield dicts2names(sent_names, sent_segs), sent_segs
        thread.join()

    def paragraphs(self):
        """Return segments for each sentence for each paragraph."""
        queue = Queue(maxsize=MAX_Q_SIZE)
        thread = Thread(target=self._parse_morph, args=(queue,))
        thread.start()
        for par_id, contents in groupby(
                queue_iterator(queue),
                key=lambda elem: elem[0]):
            contents = imap(lambda elem: elem[1:], contents)
            yield (par_id, contents)
        thread.join()

    def paragraphs_with_names(self, ann_named):
        """Return segments *and* NEs for each sentence for each paragraph."""
        queue = Queue(maxsize=MAX_Q_SIZE)
        thread = Thread(target=self._parse_names, args=(ann_named, queue))
        thread.start()
        names_iterator = queue_iterator(queue)
        for par_id, par_contents in self.paragraphs():
            par_contents = list(par_contents)
            par_names = take(len(par_contents), names_iterator)
            yield (par_id, [(sent_id, sent, dicts2names(names, sent))
                for (sent_id, sent), names
                in izip_longest(par_contents, par_names)])
        thread.join()
    
    def _parse_morph(self, queue):
        if self._gzipped:
            f = gzip.open(os.path.join(self._dirname, 'ann_morphosyntax.xml.gz'))
        else:
            f = open(os.path.join(self._dirname, 'ann_morphosyntax.xml'))
        handler = MorphHandler(queue)
        xml.sax.parse(f, handler)
        f.close()
    
    def _parse_names(self, filename, queue):
        if filename.endswith(".gz"):
            f = gzip.open(os.path.join(self._dirname, filename))
        else:
            f = open(os.path.join(self._dirname, filename))
        handler = NamesHandler(queue)
        xml.sax.parse(f, handler)
        f.close()



# encoding: utf-8

from libc.stdlib cimport malloc, free

cdef class DArray:
    def __cinit__(self):
        self.size = 0
        self.array = NULL
    def __init__(self, size):
        self.size = size
        self.array = <double*>malloc(size * sizeof(double))
    def __dealloc__(self):
        if self.array is not NULL:
            free(self.array)




# encoding: utf-8

"""
Translation between a list of labels and a list of named entities (NEs),
where list of NEs represents a collection of NEs trees.
"""

from ner.data import NamedEntity, Segment

__all__ = ["decode", "encode", "tuple2label", "label2tuple"]

_last_id = 0

def _get_next_id():
    _last_id += 1
    return 'n_%d' % _last_id

def _get_new_ne_types(ne):
    return [netype.replace('-s', '') for netype in ne if netype.endswith('-s')]

def _get_cont_ne_types(ne):
    return [netype.replace('-c', '') for netype in ne if netype.endswith('-c')]

def _recover_new_ne_types(part_res, cont_ne_types, curr_segid, max_dist,
        id2sent_pos):
    if len(cont_ne_types) == 0:
        return []
    else:
        max_nest = len(cont_ne_types) - 1
        cont_ne = _get_last_ne_of_type(part_res, cont_ne_types[0], max_nest)
        if cont_ne is None or _get_distance(part_res, cont_ne,
                curr_segid, id2sent_pos) > max_dist:
            recovered_nes = [cont_ne_types[0]]
            recovered_nes.extend(_recover_new_ne_types(part_res,
                cont_ne_types[1:], curr_segid, max_dist, id2sent_pos))
            return recovered_nes
        else:
            return []

def _get_new_nes_list(segid, new_ne_types, curr_nes_num, generate_ne_id):
    """
    Returns the list of new named entities (as maps)
    @param segid: current segment id
    @param new_ne_types: list of ne types for current segment suffixed with '-s'
    @param curr_nes_num: number of already found nes - needed in setting ne id
    """
    res = []
    next_ne_num = curr_nes_num + 1
    for ne_name in [ne_name for ne_name in reversed(new_ne_types)]:
        ne_type = ne_name.replace('-s', '')
        if '@' in ne_type:
            deriv_type = ne_type.split('@')[1]
            ne_type = ne_type.split('@')[0]
        else:
            deriv_type = None
        # ne_id = generate_ne_id(sent_id, next_ne_num)
        ne_id = generate_ne_id(next_ne_num)
        next_ne_num += 1
        ne_ptrs = []
        if len(res) != 0:
            res[-1]['ptrs'].append(ne_id)
        new_ne = {'gentype':ne_type, 'id':ne_id, 'ptrs':ne_ptrs}
        if not deriv_type is None:
            new_ne['derivType'] = deriv_type
        res.append(new_ne)

    if len(res) != 0:
        # TODO: Może potrzebna metoda konstrukcji odnosnikow ?
        # Doklejenie "ann_morph..." mozna przeprowadzic w postprocessingu.
        # res[-1]['ptrs'].append('ann_morphosyntax.xml#' + segid)
        res[-1]['ptrs'].append(segid)

    return res

#def generate_ne_id(sentid, ne_num):
#    """
#    Generates ne id with particular number
#    @param sentid: current sentence id
#    @param ne_num: number of the ne
#    """
#    # TODO !
#    return sentid.replace('morph', 'named') + '-n' + str(ne_num)

def _get_last_ne_of_type(part_res, ne_type, max_nest):
    """
    Returns the last ne in part_res with conditions specified
    @param part_res: current result
    @param ne_type: type of ne
    @param max_nest: max nest level for returned ne
    """
    for name in reversed(part_res):
        if name['gentype'] == ne_type and\
                _get_nest_level(part_res, name) <= max_nest:
            return name
    return None

def _get_nest_level(part_res, name):
    for it_name in part_res:
        if _is_child_of(name, it_name):
            return _get_nest_level(part_res, it_name) + 1
    return 0

def _is_child_of(name, parent_name):
    return name['id'] in parent_name['ptrs']

def _first_good_cont(part_res, cont_ne_types, curr_segid, max_dist,
        id2sent_pos):
    """
    Returns first suffixed with '-c' named entity 
    that is in fact continuation of previously started one
    """
    if len(cont_ne_types) == 0:
        return None
    else:
        max_nest = len(cont_ne_types) - 1
        cont_ne = _get_last_ne_of_type(part_res, cont_ne_types[0], max_nest)
        if cont_ne is None or _get_distance(part_res, cont_ne,
                curr_segid, id2sent_pos) > max_dist:
            return _first_good_cont(part_res, cont_ne_types[1:], curr_segid,
                    max_dist, id2sent_pos)
        else:
            return cont_ne

def _get_distance(part_res, ne_type, from_segid, id2sent_pos):
    """
    Returns distance from the last occurrence of ne_type in part_res
    to the segment with id==from_segid
    """
    segid1 = _get_seg_ids(ne_type, part_res)[-1]
    segid2 = from_segid
    return abs(id2sent_pos[segid1] - id2sent_pos[segid2])

def _get_seg_ids(ne, nes_list):
    """Return segment ids related to ne named entity.

    :params:
    --------
    nes_list : list of all named entities.
    seg_ids : list of ids of all segments.
    """
    res = []
    for ptr in ne['ptrs']:
        nested_ne = [ne for ne in nes_list if ne['id'] == ptr]
        if len(nested_ne) == 0:
            res.append(ptr)
        else:
            res.extend(_get_seg_ids(nested_ne[0], nes_list))
    return res

def _postprocess_ne_types(nes_list):
    for ne in nes_list:
        oldtype = ne['gentype']
        if "." in oldtype:
            basetype = oldtype.split('.')[0]
            subtype = oldtype.split('.')[1]
        else:
            subtype = None
            basetype = oldtype
        ne['type'] = basetype
        ne['subtype'] = subtype
        del ne['gentype']

def _dict2name(name_dict, entities_dict):
    return NamedEntity(
                id=name_dict['id'],
                type=name_dict['type'],
                subtype=name_dict['subtype'],
                derivType=name_dict.get('derivType', None),
                ptrs=list(map(
                              lambda id: entities_dict[id], 
                              name_dict['ptrs'])))

def _retain_only_root_names(res):
    names_to_remove = set()
    for name in res:
        for desc_name in name.get_descendant_names():
            names_to_remove.add(desc_name)
    return [name for name in res if name not in names_to_remove]

def _sort_names_and_segs_key(e, segs):
    if isinstance(e, Segment):
        return (segs.index(e), 0)
    else:
        min_seg_idx, _ = \
            min([_sort_names_and_segs_key(ptr, segs) for ptr in e.ptrs])
        return (min_seg_idx, len(e.get_descendant_names()))

def _sort_names_and_segs(names, segs):
    names.sort(key=lambda name: _sort_names_and_segs_key(name, segs))
    for name in names:
        name.ptrs.sort(key=lambda name: _sort_names_and_segs_key(name, segs))

def dicts2names(names_dicts, segs):
    res = []
    entities_dict = {}
    
    for seg in segs:
        entities_dict[seg.id] = seg
    
    for name_dict in reversed(names_dicts):
        named_entity = _dict2name(name_dict, entities_dict)
        res.insert(0, named_entity)
        entities_dict[name_dict['id']] = named_entity
    res = _retain_only_root_names(res)
    _sort_names_and_segs(res, segs)
    return res

def decode(labels, segs, generate_ne_id=lambda k: "n_" + str(k),
        recover_not_started=False, max_dist=float('+infinity')):
    """Translate a list of labels in a sentence to a list of NEs.

    :params:
    --------
    labels :
        Labels in a sentence.
    segs :
        All segments in a sentence.
    generate_ne_id :
        Method of a new named entities ids generation.
    recover_not_started :
        If True, treat not started '-c' labels as '-s'. Otherwise ignore them.
    max_dist :
        Indicates how far from the rest of the name the '-c' name parts can be.
    """
    nes_list = []
    
    seg_ids = map(lambda seg: seg.id, segs)

    id2sent_pos = dict((seg_id, i) for i, seg_id in enumerate(seg_ids))
    for segid, label in zip(seg_ids, labels):
        new_ne_types = _get_new_ne_types(label)
        cont_ne_types = _get_cont_ne_types(label)

        if recover_not_started:
            new_ne_types.extend(_recover_new_ne_types(nes_list, cont_ne_types,
                segid, max_dist, id2sent_pos))

        new_nes = _get_new_nes_list(segid, new_ne_types, len(nes_list),
                generate_ne_id)
        cont_ne = _first_good_cont(nes_list, cont_ne_types, segid, max_dist,
                id2sent_pos)
        if not cont_ne is None:
            if len(new_nes) > 0:
                cont_ne['ptrs'].append(new_nes[0]['id'])
            else:
                # cont_ne['ptrs'].append('ann_morphosyntax.xml#' + segid)
                cont_ne['ptrs'].append(segid)

        nes_list.extend(new_nes)
    _postprocess_ne_types(nes_list)
    return dicts2names(nes_list, segs)

def _name2labels_dict(name):
    upper_dict = {}
    segs = name.get_segs()
    upper_dict[segs[0]] = ['%s-s' % name.get_label()]
    lower_dict = {}
    for seg in segs[1:]:
        upper_dict[seg] = ['%s-c' % name.get_label()]
    for subname in filter(lambda ptr: isinstance(ptr, NamedEntity), name.ptrs):
        lower_dict.update(_name2labels_dict(subname))
    
    for seg, labels_list in lower_dict.iteritems():
        upper_labels_list = upper_dict[seg]
        upper_dict[seg] = labels_list + upper_labels_list
    return upper_dict
        

def encode(names, segs):
    """Translate a list of NEs a sentence to a list of labels."""
    seg2labels = {}
    for name in names:
        seg2labels.update(_name2labels_dict(name))
    return [tuple(seg2labels.get(seg, [])) for seg in segs]

def tuple2label(label):
    if len(label) == 0:
        return 'O'
    return "NE_" + '#'.join(label)

def label2tuple(lstr):
    if lstr == 'O':
        return ()
    return tuple(lstr.lstrip("NE_").split("#"))

if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)




# encoding: utf-8

from itertools import groupby, chain, product

import rawdata
import translate
from alphabet import Alphabet
from sentence import Sentence
from crf import const
import re

def argmin(l):
    i, x = min(enumerate(l), key=lambda (i, x): x)
    return i

def _alphabet_from_rdata(rdata, void_label, dummy_label):
    """Extract alphabet (of observations and labels) from
    given raw data."""
    alphabet = Alphabet(void_label=void_label, dummy_label=dummy_label)
    for (sent, labels) in rdata:
        for word in sent:
            for x in chain(*word):
                alphabet.add_observation(x)
        for y in labels:
            alphabet.add_label(y)
    return alphabet

#def _translate_rsent(rsent, alphabet):
#    """Translate raw sentence of words, where each word is a list
#    of observation values, with respect to the given alphabet."""
#    sentence = [[alphabet.observation(x) for x in word]
#                for (word, _) in rsent]
#    labels = [alphabet.label(label) for (_, label) in rsent]
#    return Sentence(sentence), labels

def _file_lines(file_name):
    f = open(file_name)
    try:
        value = map(lambda x: x.strip(), f.readlines())
    finally:
        f.close()
    return value

def labels_in_data(data):
    """Labels in data (without dummy label)"""
    return sorted(set(y for _, labels in data for y in labels))

def features_in_data_gen_absent(data):
    """Features present in data + absent features."""
    def add_features(word, prev_ys, ys, features):
        (singles, pairs) = word
        for x, y in product(singles, ys):
            features.add((-1, y, x))
        for x, y, py in product(pairs, ys, prev_ys):
            features.add((py, y, x))
    all_labels = labels_in_data(data)
    features = set()
    for sent, _ in data:
        prev_ys = [const.dummy_id]
        for word in sent:
            add_features(word, prev_ys, all_labels, features)
            prev_ys = all_labels
    return sorted(features)

def features_in_data_no_absent(data):
    """Features present in data; do not generate absent features."""
    features = set()
    for sent, labels in data:
        py = const.dummy_id
        for (singles, pairs), y in zip(sent, labels):
            for x in singles:
                features.add((-1, y, x))
            for x in pairs:
                features.add((py, y, x))
            py = y
    return sorted(features)

def features_in_data(data, generate_absent=False):
    """Return features present in the data."""
    if generate_absent:
        return features_in_data_gen_absent(data)
    else:
        return features_in_data_no_absent(data)

def obtypes(data):    
    """Return number of observations types -- maximum number
    of observations per word."""
    return max(len(ss) + len(ps) for sent, _ in data for ss, ps in sent)

def _dirdata(data, dirs):
    """Return data grouped with repsect to dirs."""
    dirdata = []
    for dirname, group in groupby(zip(dirs, data),
            key=lambda (dirname, sent): dirname):
        dirdata.append((dirname,
            map(lambda (dirname, sent): sent, group)))
    return sorted(dirdata, key=lambda (dirname, group): len(group),
            reverse=True)

def divide_data(data, dirs, n, stat=(lambda x: 1)):
    """
    Divide data to n parts with respect to dirnames --
    sentences from one directory will be in the same part.

    :param stat:
        Divide data with respect to given statistic, so that
        each part will have similar statistic value.
        Statistic value is computed in the following way:
            value = sum(stat(x) for x in part)
    """
    dirdata = _dirdata(data, dirs)
    parts = [[] for i in range(n)]
    stats = [0 for _ in parts]
    for (_, group) in dirdata:
        i = argmin(stats) 
        parts[i].extend(group)
        stats[i] += sum(stat(x) for x in group)
    return parts

def divide_in_two(data, dirs, r):
    """
    Divide data to 2 parts with respect to dirnames --
    sentences from one directory will be in the same part.

    :param r:
        Relative size of the second part (with respect
        to data size).
    """
    dirdata = _dirdata(data, dirs)
    parts = [[], []]
    for (_, group) in dirdata:
        if r < 1 and len(parts[1]) >= len(parts[0]) * (r / (1.0 - r)):
            parts[0].extend(group)
        else:
            parts[1].extend(group)
    return parts

class DataSet:

    def __init__(self, rdata, dummy_label=None, void_label=None, 
            alphabet=None, dirs=None):
        """Initialize data set.

        :params:
        --------
        void_label :
            Label, which represents void information.
            Set to None, when not related.
        dummy_label :
            Label related to a position before the first word in a sentence. 
            Should be different from each of the real labels. 
        alphabet :
            Use given alphabet to translate raw data. Ignore dummy_label
            and void_label parameters.
        """
        if alphabet is not None:
            self.alphabet = alphabet
        else:
            if dummy_label is None:
                raise ValueError("Dummy label should not be None")
            self.alphabet = _alphabet_from_rdata(rdata,
                    void_label, dummy_label)
        self.rev_alphabet = self.alphabet.reversed()
        self.data = [(Sentence(translate.sentence(rsent, self.alphabet)),
                    translate.labels(labels, self.alphabet))
                    for rsent, labels in rdata]
        self.dirs = dirs if dirs != None else range(len(rdata))

    @staticmethod
    def from_file(file_name, void_label, dummy_label, dirs_file=None):
        """Read dataset from file."""
        rdata = rawdata.from_file(file_name)
        dirs = _file_lines(dirs_file) if dirs_file != None else None
        return DataSet(rdata, void_label, dummmy_label, dirs=dirs)

    def divide(self, n, stat=(lambda x: 1)):
        """Divide dataset data to n parts."""
        return divide_data(self.data, self.dirs, n, stat=stat)

    def divide_in_two(self, r):
        """
        Divide dataset data to 2 parts.

        :param r:
            Relative size of the second part (with respect
            to data size).
        """
        return divide_in_two(self.data, self.dirs, r)

    def obtypes(self):
        """Return number of observations types -- maximum number
        of observations per word."""
        return obtypes(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, i):
        if i < 0 or i >= len(self.data):
            raise IndexError
        sent, labels = self.data[i]
        return (translate.sentence(sent, self.rev_alphabet),
                translate.labels(labels, self.rev_alphabet))

    def print_data(self, filename):
        file = open(filename, "w")
	for sent, labels in self:
            for (singles, pairs), y in zip(sent, labels):
		for x in singles:
                    assert not re.search("\s", x)
                    print >> file, x.encode('utf-8'),
                print >> file, "|",
		for x in pairs:
                    assert not re.search("\s", x)
                    print >> file, x.encode('utf-8'),
                print >> file, "|", y.encode('utf-8')
	    print >> file, ""
        file.close()



#-*- coding: utf-8 -*-

import re
from itertools import groupby

# Represents an abstract "feature".
class Feature(object):
    def value(self, sent, idx):
        raise NotImplementedError("Should use subclass")
    def prefix(self, i):
        return None

# word itself. For segment with orth='Polski' --> 'Polski'
class WORD(Feature):
    def __init__(self, lower=True):
        self.lower = lower

    def value(self, sent, idx):
        orth = sent[idx].orth
        if len(orth) == 0:
            return None
        # return "W_" + (orth.lower() if self.lower else orth)
        return orth.lower() if self.lower else orth

class UPPER_WORD(Feature):
    """Orthografic form, only if has upper character. None otherwise."""
    def value(self, sent, idx):
        orth = sent[idx].orth
        if orth.lower() == orth:
            return None
        # return "UW_" + orth
        return orth

# base form of a word. For segment with orth='Polski' --> 'Polska'
class LEMMA(Feature):
    def __init__(self, lower=True):
        self.lower = lower

    def value(self, sent, idx):
        base = sent[idx].base
        if len(base) == 0:
            return None
        # return "L_" + (base.lower() if self.lower else base)
        return base.lower() if self.lower else base

# specifies if a word starts with capital letter.
# For segment with orth='Polski' --> True
class CAPITALIZED(Feature):
    def value(self, sent, idx):
        orth = sent[idx].orth
        return "CAPITALIZED" if orth[0].isupper() else None

# pecifies if a word contains only capital letters.
# for segment with orth='Polski' --> False
# for segment with orth='POLSKI' --> True
class UPPER_CASE(Feature):
    def value(self, sent, idx):
        return "UPPER" if sent[idx].orth.isupper() else None

# returns word suffix with given lenght.
# for segment with orth='Polski' and n=3 --> 'ski'
class SUFFIX(Feature):
    def __init__(self, n, lower=True):
        self.n = n
        self.lower = lower
        
    def value(self, sent, idx):
        orth = sent[idx].orth
        if abs(self.n) < len(orth):
            suf = orth[-self.n:]
            # return "SUF_" + (suf.lower() if self.lower else suf)
            return suf.lower() if self.lower else suf
        else:
            return None

    def prefix(self, i):
        return "SUF" + str(i)

# returns word prefix with given lenght.
# for segment with orth='Polski' and n=3 --> 'Pol'
class PREFIX(Feature):
    def __init__(self, n, lower=True):
        self.n = n
        self.lower = lower

    def value(self, sent, idx):
        orth = sent[idx].orth
        if abs(self.n) < len(orth):
            pref = orth[:self.n]
            # return "PREF_" + (pref.lower() if self.lower
            #         else pref)
            return pref.lower() if self.lower else pref
        else:
            return None

    def prefix(self, i):
        return "PREF" + str(i)

class SUFFIX_PRIM(SUFFIX):
    def prefix(self, i):
        return "SUFP" + str(i)

class DUMMY_SUFFIX(Feature):
    def __init__(self, lower=True):
        self.lower = lower

    def value(self, sent, idx):
        suf = sent[idx].orth
        return suf.lower() if self.lower else suf

    def prefix(self, i):
        return "SUF" + str(i)

class DUMMY_PREFIX(Feature):
    def __init__(self, lower=True):
        self.lower = lower

    def value(self, sent, idx):
        pref = sent[idx].orth
        return pref.lower() if self.lower else pref

    def prefix(self, i):
        return "PREF" + str(i)

# returns a neighbor word.
# for segment with orth='Polski' and n=-2 in sentence 'Wracam do Polski' --> 'Wracam'
class NEIGHBOR_WORD(Feature):
    def __init__(self, n, lower=True):
        self.n = n
        self.lower = lower
        
    def value(self, sent, idx):
        my_idx = idx
        sent_len = len(sent)
        word_idx = my_idx + self.n
        # feat_pref = ("NEXTW%s_" % self.n) if self.n > 0 else\
        #             ("PREVW%s_" % -self.n)
        if word_idx in range(sent_len):
            orth = sent[word_idx].orth
            # return feat_pref + (orth.lower() if self.lower else orth)
            return orth.lower() if self.lower else orth
        else:
            return None

# returns a neighbor lemma.
# for segment with orth='Polski' and n=-2 in sentence 'Wracam do Polski' --> 'Wracać'
class NEIGHBOR_LEMMA(Feature):
    def __init__(self, n, lower=True):
        self.n = n
        self.lower = lower

    def value(self, sent, idx):
        my_idx = idx
        sent_len = len(sent)
        word_idx = my_idx + self.n
        # feat_pref = ("NEXTL%s_" % self.n) if self.n > 0 else\
        #             ("PREVL%s_" % -self.n)
        if word_idx in range(sent_len):
            base = sent[word_idx].base
            # return feat_pref + (base.lower() if self.lower else base)
            return base.lower() if self.lower else base
        else:
            return None

# returns a part of speech of a word
# for segment with orth='fajny' --> 'adj'
class POS(Feature):
    def _value(self, sent, idx):
        pos = sent[idx].pos
        # return "POS_" + pos if len(pos) > 0 else None
        if len(pos) == 0:
            raise Exception, "empty pos for word " + sent[idx].orth
        return pos
    def value(self, sent, idx):
        # return "POS_" + self._value(sent, idx)
        return self._value(sent, idx)

class POS_PAIR(POS):
    def value(self, sent, idx):
        if idx == 0:
            return None
        # return "PPR_" + (POS._value(self, sent, idx - 1) +
        #         "_" + POS._value(self, sent, idx))
        return (POS._value(self, sent, idx - 1) +
                "_" + POS._value(self, sent, idx))

class CASE(Feature):
    def value(self, sent, idx):
        case = sent[idx].case
        # return "CASE_" + case if case is not None else None
        return case if case is not None else None

class CASE_PAIR(Feature):
    def value(self, sent, idx):
        if idx == 0:
            return None
        pcase = sent[idx - 1].case
        case = sent[idx].case
        if pcase is not None and case is not None:
            # return "CPR_" + pcase + "_" + case
            return pcase + "_" + case
        return None

class POS_CASE(Feature):
    def _value(self, sent, idx):
        pos = sent[idx].pos
        case = sent[idx].case
        if pos != None:
            if case != None:
                return pos + ":" + case
            return pos
        return None
    def value(self, sent, idx):
        val = self._value(sent, idx)
        # return "PC_" + val if val else None
        return val if val else None

class POS_CASE_PAIR(POS_CASE):
    def value(self, sent, idx):
        if idx == 0:
            return None
        # return "PCPR_" + (POS_CASE._value(self, sent, idx - 1) +
        return (POS_CASE._value(self, sent, idx - 1) +
                "_" + POS_CASE._value(self, sent, idx))

# specifies if a space precedes a given word.
# for segment 'Fajny' in sentence 'Fajny samochód.' --> False
# for segment '.' in sentence 'Fajny samochód.' --> True
class NPS(Feature):
    def value(self, sent, idx):
        return sent[idx].nps

class ID(Feature):
    def value(self, sent, idx):
        return "I"

class SHAPE(Feature):
    def _charmap(self, char):
        if char.islower():
            return "l"
        if char.isupper():
            return "u"
        if char.isdigit():
            return "d"
        return "x"
    def _value(self, sent, idx):
        orth = sent[idx].orth
        return ''.join(self._charmap(char) for char in orth)
    def value(self, sent, idx):
        # return "SH_" + self._value(sent, idx)
        return self._value(sent, idx)

class PACKED_SHAPE(SHAPE):
    def _value(self, sent, idx):
        orth = sent[idx].orth
        shape = [self._charmap(char) for char in orth]
        return ''.join(k for k, g in groupby(shape))
    def value(self, sent, idx):
        # return "PSH_" + self._value(sent, idx)
        return self._value(sent, idx)

class SHAPE_PAIR(SHAPE):
    def value(self, sent, idx):
        if idx == 0:
            return None
        # return "SHPR_" + (SHAPE._value(self, sent, idx - 1) +
        return (SHAPE._value(self, sent, idx - 1) +
                "_" + SHAPE._value(self, sent, idx))

class PACKED_SHAPE_PAIR(PACKED_SHAPE):
    def value(self, sent, idx):
        if idx == 0:
            return None
        # return "PSHPR_" + (PACKED_SHAPE._value(self, sent, idx - 1) +
        return (PACKED_SHAPE._value(self, sent, idx - 1) +
                "_" + PACKED_SHAPE._value(self, sent, idx))

#def _seg2f(sent, idx, fname):
#    feature = None # cobym nie mial bledu w Eclipse
#    ran = None
#    exec("feature, ran = " + fname)
#    value = [(feature.value(sent, idx + i), i)
#            for i in ran if idx + i >= 0 and
#            idx + i < len(sent)]
#    return [str(i) + fval for (fval, i) in value if fval != None]

def _seg2f(sent, idx, feature, i):
    if idx + i >= 0 and idx + i < len(sent):
        return feature.value(sent, idx + i)
    return None

def _prepare_features(features):
    return tuple(re.sub("\s+", "_", feat) for feat in features)

def _seg2features(sent, idx, features):
    singles, pairs = [], []
    for k, (is_pair, feature, i) in enumerate(features):
        featval = _seg2f(sent, idx, feature, i)
        if featval != None:
            if feature.prefix(i) is not None:
                featval = feature.prefix(i) + "_" + featval
            else:
                featval = str(k + 1) + "_" + featval
            if is_pair:
                pairs.append(featval)
            else:
                singles.append(featval)
    return _prepare_features(singles), _prepare_features(pairs)

def sent2features(sent, features):
    words = []
    for idx in range(len(sent)):
        words.append(_seg2features(sent, idx, features))
    return words

def _file_lines(filename):
    return map(lambda x: x.strip(), open(filename).readlines())

# Should be implemented in more elegant manner...
def parse_feature_names(featnames):
    result = []
    for line in featnames:
        feature, ran = None, None
        SINGLE = False
        PAIR = True
        exec("is_pair, feature, ran = " + line)
        for i in ran:
            result.append((is_pair, feature, i))
    return result



#-*- coding: utf-8 -*-

__all__ = ["TEIAnnotator"]

import os
import shutil
import gzip
from itertools import izip
from multiprocessing import JoinableQueue
from multiprocessing import Process, Lock
from ner.nerecognizer import NERecognizer
from read import TEIReader
import write

NAMES_XML = 'ann_named.xml'
NAMES_XML_GZ = NAMES_XML+'.gz'

def get_dirs(root):
    for desc_dir, dirs, files in os.walk(root):
        if any(map(
               lambda f: f in files, 
               ['ann_morphosyntax.xml', 'ann_morphosyntax.xml.gz'])):
            yield desc_dir

class NERWorker(Process):
    """Process executing tasks from a given NER tasks queue"""
    def __init__(self, model_file, tasks):
        self.model_file = model_file
        self.tasks = tasks
        Process.__init__(self)

    def _out_name(self, output_dir, save_gzipped):
        if save_gzipped:
            return os.path.join(output_dir, NAMES_XML_GZ)
        else:
            return os.path.join(output_dir, NAMES_XML)

    def _tmp_name(self, output_dir, save_gzipped):
        if save_gzipped:
            return os.path.join(output_dir, "." + NAMES_XML_GZ)
        else:
            return os.path.join(output_dir, "." + NAMES_XML)

    def _output_for(self, fname, save_gzipped):
        # fname = self._out_name(output_dir, save_gzipped)
        if save_gzipped:
            return gzip.open(fname, 'wb')
        else:
            return open(fname, 'w')

    def _annotations(self, wypluwka_dir):
        reader = TEIReader(wypluwka_dir)
        def contents_iter(contents):
            for sent_id, sent in contents:
                names = self.ner.recognize_named_entities(sent)
                yield sent_id, names
        for par_id, contents in reader.paragraphs():
            yield (par_id, contents_iter(contents))

    def annotate_dir(self, wypluwka_dir, output_dir, save_gzipped=True):
        tmp_name = self._tmp_name(output_dir, save_gzipped)
        out = self._output_for(tmp_name, save_gzipped)
        write.write(out, self._annotations(wypluwka_dir))
        out.close()
        shutil.move(tmp_name, self._out_name(output_dir, save_gzipped))
    
    def run(self):
        self.ner = NERecognizer.load(self.model_file)
        while True:
            args, kwargs = self.tasks.get()
            try:
                self.annotate_dir(*args, **kwargs)
            except Exception, e:
                print e
            self.tasks.task_done()

class NERPool:
    """Pool of ners consuming tasks from a queue"""
    def __init__(self, model_file, processes=1):
        self.tasks = JoinableQueue(processes)
        self.workers = []
        for _ in range(processes):
            worker = NERWorker(model_file, self.tasks)
            worker.start()
            self.workers.append(worker)

    def add_task(self, *args, **kwargs):
        """Add a NER task to the queue"""
        self.tasks.put((args, kwargs))

    def wait_completion(self):
        """Wait for completion of all the tasks in the queue"""
        self.tasks.join()
        for worker in self.workers:
            worker.terminate()
        for worker in self.workers:
            worker.join()

class TEIAnnotator:
    
    def __init__(self, model_file, processes=1):
        self.nerpool = NERPool(model_file, processes)
    
    def annotate(self, 
                 wypluwka_root, 
                 output_root=None, 
                 save_gzipped=True,
                 overwrite=False, 
                 verbose=True):
        if output_root == None:
            output_root = wypluwka_root
        wypluwka_root = wypluwka_root.rstrip(os.path.sep)

        ann_named = NAMES_XML_GZ if save_gzipped else NAMES_XML
        for wypluwka_dir in get_dirs(wypluwka_root):
            output_dir = os.path.join(output_root,
                    wypluwka_dir[len(wypluwka_root)+1:])
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
            if overwrite == False and ann_named in os.listdir(output_dir):
                if verbose:
                    print "Skipping", wypluwka_dir
                continue
            if verbose:
                print "Annotating", wypluwka_dir
            self.nerpool.add_task(wypluwka_dir, output_dir,
                    save_gzipped=save_gzipped)
        self.nerpool.wait_completion()
